# Copyright 2016-2018 The NOMAD Developers Group
# Copyright 2017-2018 Lorenzo Pardini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xml.sax
import logging
import numpy as np
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from nomadcore.unit_conversion.unit_conversion import convert_unit
from nomadcore.unit_conversion import unit_conversion


class InputHandler(xml.sax.handler.ContentHandler):
    def __init__(self, backend, rgkmax):
        # self.xsType = None
        self.rgkmax = rgkmax[0]
        self.rgkmaxScr = rgkmax[0]
        # print("self.rgkmax===",self.rgkmax)
        self.backend = backend
        self.inputSectionGIndex = -1
        self.inXSInput = False
        self.fromH = unit_conversion.convert_unit_function("hartree", "J")
        self.fromB_1 = 1.88972613289*10.**10.
        self.broad = 0.01
        self.gqmax = 0.0
        self.lmaxapw = 10
        self.nemptyXS = 5
        self.nemptyScr = 0
        self.ngridqDum = [1, 1, 1]
        self.ngridkXSDum = [1, 1, 1]
        self.ngridkScrDum = [0, 0, 0]
        self.ngridq = [1,1,1]
        self.ngridkXS = [1,1,1]
        self.ngridkScr = [0,0,0]
        self.rgkmaxXs = 0.0
        self.rgkmaxscr = 0.0
        self.scissor = 0.0
        self.vkloffXS = [0.0, 0.0, 0.0]
        self.vkloffXSDum = [0.0, 0.0, 0.0]
        # self.vkloffScr = [-1.0, -1.0, -1.0]
        # self.vkloffScrDum = "-1.0 -1.0 -1.0"
        self.screening = "none"
        self.bse = False
        # self.screentype = "full"

############ BSE variables ################
        self.aresbse = True
        # self.bsetype = "singlet"
        self.lmaxdielt = 14
        self.nstlbse = [0, 0, 0, 0]
        self.nstlbseDum = [0, 0, 0, 0]
        self.nstlxas = [0, 0]
        self.nstlxasDum = [0, 0]
        self.rgkmaxBse = rgkmax[0]
        self.sciavbd = True
        self.sciavqbd = False
        self.sciavqhd = False
        self.sciavqwg = False
        self.sciavtype = "spherical"
        self.xas = False
        self.xasatom = 0
        self.xasedge = "K"
        self.xasspecies = 0
        self.tddft = False
        self.acon = "false"
        self.ahc = "false"
        self.drude = [0.0, 0.0]
        self.fxcbsesplit = 0.00001
        self.fxctype = "RPA"
        self.lmaxalda = 3
        self.nwacont = 0
        self.tetra = False
        self.tetradf = "false"
        # self.xstype = "BSE"

    def endDocument(self):
        #        pass
        #if self.freqgrid == "none":
        #    self.backend.addValue("gw_max_frequency", self.freqmax)
        #    self.backend.addValue("gw_frequency_grid_type", self.fgrid)
        #    self.backend.addValue("gw_number_of_frequencies", self.nomeg)
        #       self.backend.addValue("gw_basis_set", "mixed")
        #self.backend.addValue("gw_qp_equation_treatment", "linearization")

        if self.tetradf == "true":
            self.backend.addValue("x_exciting_xs_tetra", True)
        else:
            self.backend.addValue("x_exciting_xs_tetra", False)

        for j in range(0,3):

            self.ngridq[j] = int(self.ngridqDum[j])
            self.ngridkXS[j] = int(self.ngridkXSDum[j])
            self.vkloffXS[j] = float(self.vkloffXSDum[j])
            self.ngridkScr[j] = int(self.ngridkScrDum[j])
            # self.nstlbse[j] = int (self.nstlbseDum[j])
            # self.nstlxas[j] = int (self.nstlxasDum[j])
            # self.vkloffScr[j] = float(self.vkloffScrDum[j])

        for j in range(0,4):
            self.nstlbse[j] = int (self.nstlbseDum[j])
        for j in range(0,2):
            self.nstlxas[j] = int (self.nstlxasDum[j])

        self.backend.addValue("x_exciting_xs_ngridq", self.ngridq)
        self.backend.addValue("x_exciting_xs_ngridk", self.ngridkXS)
        self.backend.addValue("x_exciting_xs_vkloff", self.vkloffXS)
        self.backend.addValue("x_exciting_xs_screening_ngridk", self.ngridkScr)
        if self.bse == True:
            self.backend.addValue("x_exciting_xs_bse_number_of_bands", self.nstlbse)
        elif self.xas == True:
            self.backend.addValue("x_exciting_xs_bse_xasatom", self.xasatom)
            self.backend.addValue("x_exciting_xs_bse_xasedge", self.xasedge)
            self.backend.addValue("x_exciting_xs_bse_xasspecies", self.xasspecies)
            self.backend.addValue("x_exciting_xs_bse_xas_number_of_bands", self.nstlxas)
        # self.backend.addValue("x_exciting_xs_bse_xas_number_of_bands", self.nstlxas)

        # for j in range(0,4):
        #    self.nstlbse[j] = int (self.nstlbseDum[j])
        # for j in range(0,2):
        #   self.nstlxas[j] = int (self.nstlxasDum[j])
        #   self.backend.addValue("x_exciting_xs_screeninig_vkloff", self.vkloffScr)

        if self.rgkmaxXs == 0.0:
            self.backend.addValue("x_exciting_xs_rgkmax", self.rgkmax)
        else:
            self.backend.addValue("x_exciting_xs_rgkmax", float(self.rgkmaxXs))

        if self.screening != "none" and self.rgkmaxScr == 0.0:
            self.backend.addValue("x_exciting_xs_screening_rgkmax", self.rgkmax)
        else:
            self.backend.addValue("x_exciting_xs_screening_rgkmax", float(self.rgkmaxScr))

        if self.bse == True:
            if self.rgkmaxBse == 0.0:
                self.backend.addValue("x_exciting_xs_bse_rgkmax", self.rgkmax)
            else:
                self.backend.addValue("x_exciting_xs_bse_rgkmax", float(self.rgkmaxBse))

    def startElement(self, name, attrs):
        # xsType = None
        if name == "xs":
            # self.inputSectionGIndex = self.backend.openSection("section_system")
            self.inXSInput = True
            xstype = attrs.getValue('xstype')
            self.backend.addValue("x_exciting_xs_xstype", xstype)
            self.backend.addValue('x_exciting_electronic_structure_method', xstype)
            # print("xstyppe===",xstype)

            try:
                self.broad = attrs.getValue('broad')
                self.backend.addValue("x_exciting_xs_broadening", self.fromH(float(self.broad)))
            except:
                self.backend.addValue("x_exciting_xs_broadening", self.fromH(self.broad))
            try:
                self.gqmax = attrs.getValue('gqmax')
                self.backend.addValue("x_exciting_xs_gqmax", self.fromB_1*(float(self.gqmax)))
            except:
                self.backend.addValue("x_exciting_xs_gqmax", self.fromB_1*(self.gqmax))
            try:
                self.gqmax = attrs.getValue('lmaxapw')
                self.backend.addValue("x_exciting_xs_lmaxapw", int(self.lmaxapw))
            except:
                self.backend.addValue("x_exciting_xs_lmaxapw", self.lmaxapw)
            try:
                self.nemptyXS = attrs.getValue('nempty')
                self.backend.addValue("x_exciting_xs_number_of_empty_states", int(self.nemptyXS))
            except:
                self.backend.addValue("x_exciting_xs_number_of_empty_states", self.nemptyXS)
            try:
                dummy = attrs.getValue('ngridq')
                self.ngridqDum = dummy.split()
            except:
                self.ngridqDum = [1, 1, 1]
            try:
                dummy = attrs.getValue('ngridk')
                self.ngridkXSDum = dummy.split()
            except:
                self.ngridkXSDum = [1, 1, 1]
            try:
                self.rgkmaxXs = attrs.getValue('rgkmax')
            except:
                self.rgkmaxXs = 0.0
            try:
                self.scissor = attrs.getValue('scissor')
                self.backend.addValue("x_exciting_xs_scissor", self.fromH(float(self.scissor)))
            except:
                self.backend.addValue("x_exciting_xs_scissor", self.scissor)
            try:
                dummy = attrs.getValue('vkloff')
                self.vkloffXSDum = dummy.split()
            except:
                self.vkloffXSDum = [0.0, 0.0, 0.0]

        elif name == "screening":
            self.screening = "screening"
            try:
                self.nemptySrc = attrs.getValue('nempty')
                self.backend.addValue("x_exciting_xs_screening_number_of_empty_states", int(self.nemptyScr))
            except:
                self.backend.addValue("x_exciting_xs_screening_number_of_empty_states", self.nemptyScr)
            try:
                dummy = attrs.getValue('ngridk')
                self.ngridkScrDum = dummy.split()
                # print("dummo===",self.ngridkScrDum)
            except:
                self.ngridkScrDum = [0, 0, 0]
            try:
                self.rgkmaxScr = attrs.getValue('rgkmax')
            except:
                self.rgkmaxSrc = 0.0
            try:
                self.screentype = attrs.getValue('screentype')
                self.backend.addValue("x_exciting_xs_screening_type", int(self.screentype))
            except:
                self.backend.addValue("x_exciting_xs_screening_type", self.screentype)

            # self.aresbse = "true"
            # self.bsetype = "singlet"
            # self.lmaxdielt = 14
            # self.nstlbse = [0, 0, 0, 0]
            # self.nstlxas = [0, 0]
            # self.rgkmaxBse = rgkmax[0]
            # self.sciavbd = "true"
            # self.sciavqbd = "false"
            # self.sciavqhd = "false"
            # self.sciavqwg = "false"
            # self.sciavtype = "spherical"
            # self.xas = "false"
            # self.xasatom = 0
            # self.xasedge = "K"
            # self.xasspecies = 0

        elif name == "BSE":
            self.bse = True
            # xsType = "BSE"
            try:
                self.aresbse = attrs.getValue('aresbse')
                if self.aresbse == "true":
                    self.backend.addValue("x_exciting_xs_bse_antiresonant", True)
                else:
                    self.backend.addValue("x_exciting_xs_bse_antiresonant", False)
            except:
                self.backend.addValue("x_exciting_xs_bse_antiresonant", True)
            try:
                self.lmaxdielt = attrs.getValue('lmaxdielt')
                self.backend.addValue("x_exciting_xs_bse_angular_momentum_cutoff", int(self.lmaxdielt))
            except:
                self.backend.addValue("x_exciting_xs_bse_angular_momentum_cutoff", self.lmaxdielt)
            try:
                self.rgkmaxBse = attrs.getValue('rgkmax')
                # self.backend.addValue("x_exciting_xs_bse_rgkmax", float(self.rgkmax))
            except:
                pass
                # self.backend.addValue("x_exciting_xs_bse_rgkmax", self.rgkmax)
            try:
                self.sciavbd = attrs.getValue('sciavbd')
                if self.sciavqbd == "true":
                    self.backend.addValue("x_exciting_xs_bse_sciavbd", True)
                else:
                    self.backend.addValue("x_exciting_xs_bse_sciavbd", False)
            except:
                self.backend.addValue("x_exciting_xs_bse_sciavbd", True)
            try:
                self.sciavqbd = attrs.getValue('sciavqbd')
                if self.sciavqbd == "true":
                    self.backend.addValue("x_exciting_xs_bse_sciavqbd", True)
                else:
                    self.backend.addValue("x_exciting_xs_bse_sciavqbd", False)
            except:
                self.backend.addValue("x_exciting_xs_bse_sciavqbd", False)
            try:
                self.sciavqhd = attrs.getValue('sciavqhd')
                if self.sciavqhd == "true":
                    self.backend.addValue("x_exciting_xs_bse_sciavqhd", True)
                else:
                    self.backend.addValue("x_exciting_xs_bse_sciavqhd", False)
            except:
                self.backend.addValue("x_exciting_xs_bse_sciavqhd", False)
            try:
                self.sciavqwg = attrs.getValue('sciavqwg')
                if self.sciavqwg == "true":
                    self.backend.addValue("x_exciting_xs_bse_sciavqwg", True)
                else:
                    self.backend.addValue("x_exciting_xs_bse_sciavqwg", False)
            except:
                self.backend.addValue("x_exciting_xs_bse_sciavqwg", False)
            try:
                self.sciavtype = attrs.getValue('sciavtype')
                self.backend.addValue("x_exciting_xs_bse_sciavtype", self.sciavtype)
            except:
                self.backend.addValue("x_exciting_xs_bse_sciavtype", self.sciavtype)
            try:
                self.xas = attrs.getValue('xas')
                if self.xas == "false":
                    # print("xas===",self.xas)
                    self.backend.addValue("x_exciting_xs_bse_xas", False)
                else:
                    # print("xas===",self.xas)
                    self.backend.addValue("x_exciting_xs_bse_xas", True)
            except:
                # print("xasdefault===",self.xas)
                self.backend.addValue("x_exciting_xs_bse_xas", self.xas)
            try:
                self.xasatom = int(attrs.getValue('xasatom'))
                # self.backend.addValue("x_exciting_xs_bse_xasatom", self.xasatom)
            except:
                pass
                # self.backend.addValue("x_exciting_xs_bse_xasatom", self.xasatom)
            try:
                self.xasedge = attrs.getValue('xasedge')
                # self.backend.addValue("x_exciting_xs_bse_xasedge", self.xasedge)
            except:
                pass
                # self.backend.addValue("x_exciting_xs_bse_xasedge", self.xasedge)
            try:
                self.xasspecies = int(attrs.getValue('xasspecies'))
                # self.backend.addValue("x_exciting_xs_bse_xasspecies", self.xasspecies)
            except:
                pass
                # self.backend.addValue("x_exciting_xs_bse_xasspecies", self.xasspecies)
                # if self.xas == False:
            try:
                dummy = attrs.getValue('nstlbse')
                self.nstlbseDum = dummy.split()
            except:
                self.nstlbseDum = [0, 0, 0, 0]
            try:
                dummy = attrs.getValue('nstlxas')
                self.nstlxasDum = dummy.split()
            except:
                self.nstlxasDum = [0, 0]

        elif name == "tddft":
            self.tddft = True

            try:
                self.acon = attrs.getValue('acont')
                if self.acon == "true":
                    self.backend.addValue("x_exciting_xs_tddft_analytic_continuation", True)
                else:
                    self.backend.addValue("x_exciting_xs_tddft_analytic_continuation", False)
            except:
                    self.backend.addValue("x_exciting_xs_tddft_analytic_continuation", False)
            try:
                self.ahc = attrs.getValue('ahc')
                if self.ahc == "true":
                    self.backend.addValue("x_exciting_xs_tddft_anomalous_Hall_conductivity", True)
                else:
                    self.backend.addValue("x_exciting_xs_tddft_anomalous_Hall_conductivity", False)
            except:
                    self.backend.addValue("x_exciting_xs_tddft_anomalous_Hall_conductivity", False)
            try:
                self.aresdf = attrs.getValue('aresdf')
                if self.aresdf == "true":
                    self.backend.addValue("x_exciting_xs_tddft_anti_resonant_dielectric", True)
                elif self.aresdf == "false":
                    self.backend.addValue("x_exciting_xs_tddft_anti_resonant_dielectric", False)
            except:
                    self.backend.addValue("x_exciting_xs_tddft_anti_resonant_dielectric", True)
            try:
                self.aresfxc = attrs.getValue('aresfxc')
                if self.aresfxc == "true":
                    self.backend.addValue("x_exciting_xs_tddft_anti_resonant_xc_kernel", True)
                elif self.aresfxc == "false":
                    self.backend.addValue("x_exciting_xs_tddft_anti_resonant_xc_kernel", False)
            except:
                    self.backend.addValue("x_exciting_xs_tddft_anti_resonant_xc_kernel", True)
            try:
                dummy = attrs.getValue('drude')
                self.drude = dummy.split()
                for j in range (0,2):
                    self.drude[j] = float(self.drude[j])
                self.backend.addValue("x_exciting_xs_tddft_drude", self.drude)
            except:
                self.backend.addValue("x_exciting_xs_tddft_drude", self.drude)
            try:
                self.fxcbsesplit = attrs.getValue('fxcbsesplit')
                self.backend.addValue("x_exciting_xs_tddft_split_parameter", self.fromH(float(self.fxcbsesplit)))
            except:
                self.backend.addValue("x_exciting_xs_tddft_split_parameter", self.fromH(self.fxcbsesplit))
            try:
                self.fxctype = attrs.getValue('fxctype')
                self.backend.addValue("x_exciting_xs_tddft_xc_kernel", self.fxctype)
            except:
                self.backend.addValue("x_exciting_xs_tddft_xc_kernel", self.fxctype)
            try:
                intraband = attrs.getValue('intraband')
                if intraband == "true":
                    self.backend.addValue("x_exciting_xs_tddft_finite_q_intraband_contribution", True)
                else:
                    self.backend.addValue("x_exciting_xs_tddft_finite_q_intraband_contribution", False)
            except:
                    self.backend.addValue("x_exciting_xs_tddft_finite_q_intraband_contribution", False)
            try:
                kerndiag = attrs.getValue('kerndiag')
                if kerndiag == "true":
                    self.backend.addValue("x_exciting_xs_tddft_diagonal_xc_kernel", True)
                else:
                    self.backend.addValue("x_exciting_xs_tddft_diagonal_xc_kernel", False)
            except:
                    self.backend.addValue("x_exciting_xs_tddft_diagonal_xc_kernel", False)
            try:
                self.lmaxalda = attrs.getValue('lmaxalda')
                self.backend.addValue("x_exciting_xs_tddft_lmax_alda", int(self.lmaxalda))
            except:
                self.backend.addValue("x_exciting_xs_tddft_lmax_alda", self.lmaxalda)
            try:
                mdfqtype = attrs.getValue('mdfqtype')
                if mdfqtype == "0":
                    self.backend.addValue("x_exciting_xs_tddft_macroscopic_dielectric_function_q_treatment", int(mdfqtype))
                else:
                    self.backend.addValue("x_exciting_xs_tddft_macroscopic_dielectric_function_q_treatment", int(mdfqtype))
            except:
                    mdfqtype = 0
                    self.backend.addValue("x_exciting_xs_tddft_macroscopic_dielectric_function_q_treatment", int(mdfqtype))
            try:
                self.nwacont = attrs.getValue('nwacont')
                self.backend.addValue("x_exciting_xs_tddft_analytic_continuation_number_of_intervals", int(self.nwacont))
            except:
                self.backend.addValue("x_exciting_xs_tddft_analytic_continuation_number_of_intervals", self.nwacont)

        elif name == "tetra":
            self.tetra = True
            try:
                self.tetradf = attrs.getValue('tetradf')
                # if self.tetradf == "true":
                #   selfbackend.addValue("x_exciting_xs_tetra", True)
                # else:
                #   self.backend.addValue("x_exciting_xs_tetra", False)
            except:
                self.tetradf == "false"
                # self.backend.addValue("x_exciting_xs_tetra", False)

    def endElement(self, name):
        pass

def parseInput(inF, backend, gmaxvr):
    handler = InputHandler(backend, gmaxvr)
    logging.info("will parse")
    xml.sax.parse(inF, handler)
    logging.info("did parse")



# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
from cp2kparser.versions.cp2k262.singlepointforceparser import CP2KSinglePointForceParser
from nomadcore.caching_backend import CachingLevel
from cp2kparser.versions.cp2k262.commonparser import CP2KCommonParser
import logging
logger = logging.getLogger("nomad")


class CP2KSinglePointParser(MainHierarchicalParser):
    """The main parser class. Used to parse the CP2K calculation with run types:
        -ENERGY
        -ENERGY_FORCE
    """
    def __init__(self, parser_context):
        """
        """
        super(CP2KSinglePointParser, self).__init__(parser_context)
        self.setup_common_matcher(CP2KCommonParser(parser_context))

        #=======================================================================
        # Cache levels
        self.caching_levels.update({
            'x_cp2k_energy_total_scf_iteration': CachingLevel.ForwardAndCache,
            'x_cp2k_energy_XC_scf_iteration': CachingLevel.ForwardAndCache,
            'x_cp2k_energy_change_scf_iteration': CachingLevel.ForwardAndCache,
            'x_cp2k_stress_tensor': CachingLevel.ForwardAndCache,
            'x_cp2k_section_stress_tensor': CachingLevel.ForwardAndCache,
        })

        #=======================================================================
        # SimpleMatchers
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=['section_run', "section_single_configuration_calculation", "section_system", "section_method"],
            otherMetaInfo=["atom_forces"],
            subMatchers=[
                self.cm.header(),
                self.cm.quickstep_header(),
                self.cm.quickstep_calculation(),
                self.cm.footer(),
            ]
        )

    #===========================================================================
    # onClose triggers
    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        """
        """
        # If the force file for a single point calculation is available, and
        # the forces were not parsed fro the output file, parse the separate
        # file
        if section["atom_forces"] is None:
            force_file = self.file_service.get_file_by_id("force_file_single_point")
            if force_file is not None:
                force_parser = CP2KSinglePointForceParser(self.parser_context)
                force_parser.parse(force_file)
            else:
                logger.warning("The file containing the forces printed by ENERGY_FORCE calculation could not be found.")

        # Only in the single configuration calculations the number of scf
        # iterations is given. E.g. in geometry optimization there are multiple
        # scf calculations so this loses it's meaning sort of.
        self.cache_service.addValue("number_of_scf_iterations")

    def onClose_x_cp2k_section_scf_iteration(self, backend, gIndex, section):
        """Keep track of how many SCF iteration are made."""
        self.cache_service["number_of_scf_iterations"] += 1
        gId = backend.openSection("section_scf_iteration")
        section.add_latest_value("x_cp2k_energy_total_scf_iteration", "energy_total_scf_iteration")
        section.add_latest_value("x_cp2k_energy_XC_scf_iteration", "energy_XC_scf_iteration")
        section.add_latest_value("x_cp2k_energy_change_scf_iteration", "energy_change_scf_iteration")
        backend.closeSection("section_scf_iteration", gId)

    def onClose_x_cp2k_section_quickstep_calculation(self, backend, gIndex, section):
        """"""
        section.add_latest_value("x_cp2k_energy_total", "energy_total")
        section.add_latest_value("x_cp2k_electronic_kinetic_energy", "electronic_kinetic_energy")
        section.add_latest_value("x_cp2k_quickstep_converged", "single_configuration_calculation_converged")
        section.add_latest_array_values("x_cp2k_atom_forces", "atom_forces")

    def onClose_x_cp2k_section_stress_tensor(self, backend, gIndex, section):
        """"""
        gId = backend.openSection("section_stress_tensor")
        section.add_latest_array_values("x_cp2k_stress_tensor", "stress_tensor")
        backend.closeSection("section_stress_tensor", gId)

    def onClose_section_system(self, backend, gIndex, section):
        """Stores the index of the section method. Should always be 0, but
        let's get it dynamically just in case there's something wrong.
        """
        self.cache_service.addArrayValues("atom_positions", unit="angstrom")
        self.cache_service.addArrayValues("simulation_cell", unit="angstrom")
        self.cache_service.addArrayValues("lattice_vectors", unit="angstrom")

    #===========================================================================
    # adHoc functions



# Copyright 2015-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa, Ask Hjorth Larsen, Jens Jørgen Mortensen
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
import re
#p = re.compile(r"(?P<x_name>.*_X.*)\+(?P<c_name>.*_C.*)")
p = re.compile(
  '((?P<x_name>(GGA|LDA|MGGA|HF|HYB_MGGA)_X.*)|(?P<c_name>(GGA|LDA|MGGA)_C.*))')

short_names = {
        'LDA': 'LDA_X+LDA_C_PW',
        'PW91': 'GGA_X_PW91+GGA_C_PW91',
        'PBE': 'GGA_X_PBE+GGA_C_PBE',
        'PBEsol': 'GGA_X_PBE_SOL+GGA_C_PBE_SOL',
        'revPBE': 'GGA_X_PBE_R+GGA_C_PBE',
        'RPBE': 'GGA_X_RPBE+GGA_C_PBE',
        'BLYP': 'GGA_X_B88+GGA_C_LYP',
        'HCTH407': 'GGA_XC_HCTH_407',
        'WC': 'GGA_X_WC+GGA_C_PBE',
        'AM05': 'GGA_X_AM05+GGA_C_AM05',
        # 'M06-L': 'MGGA_X_M06_L+MGGA_C_M06_L',
        # 'TPSS': 'MGGA_X_TPSS+MGGA_C_TPSS',
        # 'revTPSS': 'MGGA_X_REVTPSS+MGGA_C_REVTPSS',
        'mBEEF': 'MGGA_X_MBEEF+GGA_C_PBE_SOL'}


def get_libxc_name(name):
    if name in short_names:
        libxc_name = short_names[name]
    else:
        libxc_name = name
    return libxc_name

def get_libxc_xc_names(name):
    name = get_libxc_name(name)
    xc = {'xc_name': None,
          'x_name' : None,
          'c_name': None}

    if '_XC_' in name:
        xc['xc_name'] = name
        return xc

    if '+' in name:
        s = name.split('+')
        xc['x_name'] = s[0]
        xc['c_name'] = s[1]
        return xc

    m = re.search(p, name)
    if m is not None: # it is either a correlation or exchange functional
        xc.update(m.groupdict())
        return xc

    xc['xc_name'] = name  # for something like BEEF-vdW
    return xc

if __name__ == '__main__':
#    print(get_libxc_name('LDA'))
#    print(get_libxc_name('GGA_X_PBE'))
    names = ['GGA_X_B88+GGA_C_LYP',
             'HF_X',
             'HYB_GGA_XC_B1LYP',
             'HYB_GGA_XC_HSE03',
             'BEEF-vdW',
             'LDA_K_TF']
    for name in names:
        print(get_libxc_xc_names(name))



import requests
from requests.exceptions import Timeout, TooManyRedirects, RequestException, HTTPError
from datetime import datetime
import re
import sys
import numpy as np
import json
from tinydb import TinyDB, Query

dbasefile = "openkim_query_database.json"

kimdb = TinyDB(dbasefile)

atomlabel = 'Al'
structure = 'fcc'

#nomad_space_group_map={
#    'sc'     : '221',
#    "fcc"    : '225',
#    'bcc'    : '229',
#    'diamond': '227'
#}

def secondsFromEpoch(date):
    epoch = datetime(1970,1,1)
    ts=date-epoch
    return ts.seconds + ts.microseconds/1000.0

def get_timestep(kim_date):
    pdate = None
    if kim_date:
        pdate = datetime.strptime(kim_date.strip(), "%Y-%m-%d %H:%M:%S.%f")
    if pdate:
        pdate = secondsFromEpoch(pdate)
    return pdate

def OPENKIM_query(atomlabel, structure, properties=None):
    """
    atomlabel:
    ----------
    string : element symbol
    
    structure:
    ----------
    string : fcc, bcc, sc, diamond

    properties:
    -----------
    string : lattice_energy, elastic_constants

    returns:
    --------
    dictionary of OpenKIM entry
    """
    openkim_query = None
    try:
        query = requests.post(
                url="https://query.openkim.org/api",
                data={
                    'flat'    : 'on',
                    'database': 'data',
                    'limit'   : '0',
                    'fields'  : json.dumps({
                        #"a.si-value": "1",
                        #"cohesive-potential-energy.si-value" : "1",
                        #"meta.subject.kimcode": "1"
                        }),
                    'query' : json.dumps({
                        "meta.type"   : "tr",
                    #"property-id" : {
                    #    "$regex"  : ":property/structure-cubic-crystal-npt"
                    #},
                    #"meta.runner.kimcode" : {
                    #    "$regex"  : "^LatticeConstantCubicEnergy"
                    #},
                    #"species.source-value": {
                    #    "$all" : [atomlabel], 
                    #    "$not" : {
                    #        "$elemMatch" : {
                    #            "$nin" : [atomlabel]
                    #            }
                    #        }
                    #    },
                    "short-name.source-value": structure
                    })
                }
            )
        query.raise_for_status()
        openkim_query = query.json()
    except(ConnectionError, HTTPError, Timeout, TooManyRedirects, RequestException):
        print("OpenKIM request exception: %s" % sys.exc_info()[1])

    return openkim_query
   
openkim_data  = OPENKIM_query(atomlabel, structure)
if openkim_data:
    
    now = datetime.today().strftime("%Y-%m-%d %H:%M:%S.%f")

    updated_data = []
    kim_item = Query()
    for item in openkim_data:
        kim_instance_id = item["instance-id"]
        kim_unique_code = item["meta._id"]
        kim_result_code = item["meta.test-result-id"]
        kim_unique_code = kim_unique_code + '-' + str(kim_instance_id)
        kim_result_code = kim_result_code + '-' + str(kim_instance_id)
        kim_create_date = get_timestep(item["meta.created_on"])
        kim_insert_date = get_timestep(item["inserted_on"])
        db_entry = kimdb.search((kim_item.unique_id == kim_unique_code and 
            kim_item.result_id == kim_result_code))
        if db_entry:
            if db_entry[0]["insert_timestep"] < kim_insert_date:
                kimdb.update({
                    'insert_timestep': kim_insert_date, 
                    }, 
                    (kim_item.unique_id == kim_unique_code and
                     kim_item.result_id == kim_result_code))
                updated_data.append(item)
        else:
            kimdb.insert({
                'unique_id': kim_unique_code, 
                'result_id': kim_unique_code, 
                'insert_timestep': kim_insert_date, 
                })
            updated_data.append(item)
   
    if updated_data:
        openkim_query = {}
        openkim_query['OPENKIM_QUERY_OUTPUT'] = 'OPENKIM_QUERY_OUTPUT'
        openkim_query['QUERY'] = updated_data

        with open('data.json', 'w') as outfile:
            json.dump(openkim_query, outfile, 
                    sort_keys = True, indent = 4,
                    ensure_ascii = True)
        with open('data.json', 'r') as databfile:
            qdata = json.load(databfile)




# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module constructs the parser for the RIDFT module from TurboMole"""

import logging
import re
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import RE_FLOAT
import turbomoleparser.TurbomoleCommon as Common

logger = logging.getLogger("nomad.turbomoleParser")


class RIDFTparser(object):

    def __init__(self, context, key="ridft"):
        context[key] = self
        self.__context = context
        self.__backend = None

    def purge_data(self):
        pass

    def set_backend(self, backend):
        self.__backend = backend

    def build_parser(self):
        references = SM(r"\s{5,}[^+ ].+$",
                        name="references",
                        coverageIgnore=True,
                        repeats=True,
                        )
        header = SM(r"\s*r i d f t\s*$",
                    name="Credits",
                    coverageIgnore=True,
                    subMatchers=[references],
                    endReStr=r"\s*\+-+\+"
                    )

        auxbasis_title = r"\s*RI-J\s+AUXILIARY\s+BASIS\s+SET\s+information\s*:\s*$"

        sub_matchers = [
            header,
            self.__context["method"].build_uhf_matcher(),
            self.__context["geo"].build_qm_geometry_matcher(),
            self.__context["geo"].build_orbital_basis_matcher(),
            self.__context["geo"].build_auxiliary_basis_matcher(),
            # TODO: verify if this auxbasis title is used in official releases
            self.__context["geo"].build_auxiliary_basis_matcher(auxbasis_title),
            self.__context["orbitals"].build_ir_rep_matcher(),
            self.__context["method"].build_dft_functional_matcher(),
            self.__context["embedding"].build_embedding_matcher(),
            self.__build_scf_cycle_matcher(),
            self.__context["method"].build_total_energy_matcher(),
            self.__context["orbitals"].build_eigenstate_matcher(),
        ]

        return self.__context.build_module_matcher("ridft", sub_matchers, "RIDFT",
                                                   self.__context["method"].add_default_functional)

    def __build_scf_cycle_matcher(self):

        class PreviousCycle(object):
            energy = None

        def compute_energy_difference(backend, groups):
            backend.addRealValue("energy_total_scf_iteration", float(groups[0]), unit="hartree")
            backend.addRealValue("x_turbomole_energy_1electron_scf_iteration",
                                 float(groups[1]), unit="hartree")
            backend.addRealValue("x_turbomole_energy_2electron_scf_iteration",
                                 float(groups[2]), unit="hartree")
            # TODO: clarify the meaning of the "NORM[dD(SAO)]" and "TOL" columns in the output
            if PreviousCycle.energy:
                backend.addRealValue("energy_change_scf_iteration",
                                 float(groups[0]) - PreviousCycle.energy, unit="hartree")
            PreviousCycle.energy = float(groups[0])

        def store_damping(backend, groups):
            backend.addRealValue("x_turbomole_damping_scf_iteration",
                                 1.0 / (1.0 + float(groups[0])))

        total_energy = SM(r"\s*[0-9]+\s+("+RE_FLOAT+")\s+("+RE_FLOAT+")\s+("+RE_FLOAT+")"
                          "\s+("+RE_FLOAT+")\s+("+RE_FLOAT+")",
                          name="SCF E total",
                          startReAction=compute_energy_difference,
                          required=True
                          )

        xc_energy = SM(r"\s*Exc =\s*(?P<energy_XC_scf_iteration__hartree>"+RE_FLOAT+")"
                       r"\s+Coul =\s*(?P<energy_electrostatic_scf_iteration__hartree>"+RE_FLOAT+")",
                       name="SCF E xc+coul",
                       )
        damping = SM(r"\s*current damping\s*=\s*("+RE_FLOAT+")\s*$",
                     name="SCF damping",
                     startReAction=store_damping
                     )

        # TODO: identify the units of these norms
        norm_diis = SM(r"\s*Norm of current diis error:\s*(?P<x_turbomole_norm_diis_scf_iteration>"
                       + RE_FLOAT+r")\s*$",
                       name="SCF norm DIIS"
                       )
        norm_fia_block = SM(r"\s*max. resid. norm for Fia\-block\s*=\s*(?P"
                            r"<x_turbomole_norm_fia_scf_iteration>"+RE_FLOAT+")\s*for orbital\s+"
                            r"(?P<x_turbomole_norm_fia_orbital_scf_iteration>[0-9]+[a-z][0-9'\"]?"
                            r"\s*(?:alpha|beta)?)\s*$",
                            name="SCF norm Fia"
                            )
        norm_fock = SM(r"\s*max. resid. fock norm\s*=\s*(?P"
                       r"<x_turbomole_norm_fock_scf_iteration>"+RE_FLOAT+")\s*for orbital\s+"
                       r"(?P<x_turbomole_norm_fock_orbital_scf_iteration>[0-9]+[a-z][0-9'\"]?"
                       r"\s*(?:alpha|beta)?)\s*$",
                       name="SCF norm Fock"
                       )

        scf_iteration = SM("\s*ITERATION\s+ENERGY\s+1e-ENERGY\s+2e-ENERGY\s+"
                           "RMS\[dD\(SAO\)\]\s+TOL",
                           name="SCF iteration",
                           repeats=True,
                           sections=["section_scf_iteration"],
                           subMatchers=[
                               total_energy,
                               xc_energy,
                               damping,
                               norm_diis,
                               norm_fia_block,
                               norm_fock
                           ]
                           )

        return SM(r"\s*Starting SCF iterations\s*$",
                  name="HF/DFT SCF",
                  required=True,
                  subMatchers=[
                      scf_iteration
                  ]
                  )



# Copyright 2018 Markus Scheidgen, Alvin Noe Ladines
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import requests
import click
from bravado import requests_client as bravado_requests_client
from bravado import client as bravado_client
from urllib import parse as urllib_parse

from nomad import config as nomad_config
from nomad import utils
from nomad import client as nomad_client

from nomad.cli.cli import cli


def create_client():
    return _create_client()


def _create_client(*args, **kwargs):
    return __create_client(*args, **kwargs)


def __create_client(
        user: str = nomad_config.client.user,
        password: str = nomad_config.client.password,
        ssl_verify: bool = True, use_token: bool = True):
    ''' A factory method to create the client. '''
    if not ssl_verify:
        import warnings
        warnings.filterwarnings("ignore")

    http_client = bravado_requests_client.RequestsClient(ssl_verify=ssl_verify)

    client = bravado_client.SwaggerClient.from_url(
        '%s/swagger.json' % nomad_config.client.url,
        http_client=http_client)

    utils.get_logger(__name__).info('created bravado client', user=user)

    if user is not None:
        host = urllib_parse.urlparse(nomad_config.client.url).netloc.split(':')[0]
        if use_token:
            http_client.authenticator = nomad_client.KeycloakAuthenticator(
                host=host,
                user=user,
                password=password,
                server_url=nomad_config.keycloak.server_url,
                realm_name=nomad_config.keycloak.realm_name,
                client_id=nomad_config.keycloak.client_id)
        else:
            http_client.set_basic_auth(
                host=host,
                username=user,
                password=password)

        utils.get_logger(__name__).info('set bravado client authentication', user=user)

    return client


def handle_common_errors(func):
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except requests.exceptions.ConnectionError:
            click.echo(
                '\nCould not connect to nomad at %s. '
                'Check connection and url.' % nomad_config.client.url)
            sys.exit(0)
    return wrapper


@cli.group(help='Commands that use the nomad API to do useful things')
@click.option('-n', '--url', default=nomad_config.client.url, help='The URL where nomad is running, default is "%s".' % nomad_config.client.url)
@click.option('-u', '--user', default=None, help='the user name to login, default is "%s" login.' % nomad_config.client.user)
@click.option('-w', '--password', default=nomad_config.client.password, help='the password used to login.')
@click.option('--no-ssl-verify', help='disables SSL verificaton when talking to nomad.', is_flag=True)
@click.option('--no-token', is_flag=True, help='replaces token with basic auth, e.g. to work with v0.6.x or older API versions')
@click.pass_context
def client(ctx, url: str, user: str, password: str, no_ssl_verify: bool, no_token: bool):
    logger = utils.get_logger(__name__)

    logger.info('Used nomad is %s' % url)
    logger.info('Used user is %s' % user)

    nomad_config.client.url = url

    ctx.obj.user = user

    global _create_client

    def _create_client(*args, **kwargs):  # pylint: disable=W0612
        if user is not None:
            logger.info('create client', user=user)
            return __create_client(
                user=user, password=password, ssl_verify=not no_ssl_verify,
                use_token=not no_token)
        else:
            logger.info('create anonymous client')
            return __create_client(ssl_verify=not no_ssl_verify)



from ase.data import chemical_symbols
from elasticsearch_dsl import Keyword, Float, InnerDoc, Nested, Integer
import numpy as np

from nomad.units import ureg
from nomad.metainfo import MSection, Section, Quantity, SubSection, MEnum, DefinitionAnnotation
from nomad.metainfo.search_extension import Search


# TODO move the module

def optimade_links(section: str):
    return [
        'https://github.com/Materials-Consortia/OPTiMaDe/blob/develop/optimade.md#%s' %
        section]


class ElementRatio(InnerDoc):
    elements = Keyword()
    elements_ratios = Float()

    @staticmethod
    def from_structure_entry(entry: 'OptimadeEntry'):
        return [
            ElementRatio(elements=entry.elements[i], elements_ratios=entry.elements_ratios[i])
            for i in range(0, entry.nelements)]


class Optimade(DefinitionAnnotation):
    def __init__(self, query: bool = False, entry: bool = False):
        pass


class Species(MSection):
    '''
    Used to describe the species of the sites of this structure. Species can be pure
    chemical elements, or virtual-crystal atoms representing a statistical occupation of a
    given site by multiple chemical elements.
    '''

    m_def = Section(links=optimade_links('h.6.2.13'))

    name = Quantity(
        type=str, a_optimade=Optimade(entry=True), description='''
            The name of the species; the name value MUST be unique in the species list.
        ''')

    chemical_symbols = Quantity(
        type=MEnum(chemical_symbols + ['x', 'vacancy']), shape=['1..*'],
        a_optimade=Optimade(entry=True), description='''
            A list of strings of all chemical elements composing this species.

            It MUST be one of the following:

            - a valid chemical-element name, or
            - the special value "X" to represent a non-chemical element, or
            - the special value "vacancy" to represent that this site has a non-zero probability
            of having a vacancy (the respective probability is indicated in the concentration
            list, see below).

            If any one entry in the species list has a chemical_symbols list that is longer than 1
            element, the correct flag MUST be set in the list structure_features (see
            structure_features)
        ''')

    concentration = Quantity(
        type=float, shape=['1..*'],
        a_optimade=Optimade(entry=True), description='''
            A list of floats, with same length as chemical_symbols. The numbers represent the
            relative concentration of the corresponding chemical symbol in this species. The
            numbers SHOULD sum to one. Cases in which the numbers do not sum to one typically fall
            only in the following two categories:

            - Numerical errors when representing float numbers in fixed precision, e.g. for two
            chemical symbols with concentrations 1/3 and 2/3, the concentration might look
            something like [0.33333333333, 0.66666666666]. If the client is aware that the sum
            is not one because of numerical precision, it can renormalize the values so that the
            sum is exactly one.
            - Experimental errors in the data present in the database. In this case, it is the
            responsibility of the client to decide how to process the data.

            Note that concentrations are uncorrelated between different sites (even of the same
            species).
        ''')

    mass = Quantity(type=float, unit=ureg.amu, a_optimade=dict(entry='optional'))

    original_name = Quantity(type=str, a_optimade=dict(entry='optional'), description='''
        Can be any valid Unicode string, and SHOULD contain (if specified) the name of the
        species that is used internally in the source database.

        Note: With regards to "source database", we refer to the immediate source being
        queried via the OPTiMaDe API implementation. The main use of this field is for source
        databases that use species names, containing characters that are not allowed (see
        description of the species_at_sites list).
        ''')


class OptimadeEntry(MSection):
    m_def = Section(
        links=optimade_links('h.6.2'),
        a_flask=dict(skip_none=True))

    elements = Quantity(
        type=MEnum(chemical_symbols), shape=['1..*'],
        links=optimade_links('h.6.2.1'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            Names of the different elements present in the structure.
        ''')

    nelements = Quantity(
        type=int, default=0,
        links=optimade_links('h.6.2.2'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            Number of different elements in the structure as an integer.
        ''')

    elements_ratios = Quantity(
        type=float, shape=['nelements'],
        links=optimade_links('h.6.2.3'),
        a_search=Search(mapping=Nested(ElementRatio), value=ElementRatio.from_structure_entry),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            Relative proportions of different elements in the structure.
        ''')

    chemical_formula_descriptive = Quantity(
        type=str,
        links=optimade_links('h.6.2.4'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            The chemical formula for a structure as a string in a form chosen by the API
            implementation.
        ''')

    chemical_formula_reduced = Quantity(
        type=str,
        links=optimade_links('h.6.2.5'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            The reduced chemical formula for a structure as a string with element symbols and
            integer chemical proportion numbers. The proportion number MUST be omitted if it is 1.
        ''')

    chemical_formula_hill = Quantity(
        type=str,
        links=optimade_links('h.6.2.6'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=False),
        description='''
            The chemical formula for a structure in Hill form with element symbols followed by
            integer chemical proportion numbers. The proportion number MUST be omitted if it is 1.
        ''')

    chemical_formula_anonymous = Quantity(
        type=str,
        links=optimade_links('h.6.2.7'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            The anonymous formula is the chemical_formula_reduced, but where the elements are
            instead first ordered by their chemical proportion number, and then, in order left to
            right, replaced by anonymous symbols A, B, C, ..., Z, Aa, Ba, ..., Za, Ab, Bb, ... and
            so on.
        ''')

    dimension_types = Quantity(
        type=int, shape=[3], default=[0, 0, 0],
        links=optimade_links('h.6.2.8'),
        a_search=Search(value=lambda a: sum(a.dimension_types), mapping=Integer()),
        a_optimade=Optimade(query=True, entry=True),
        description='''
            List of three integers. For each of the three directions indicated by the three lattice
            vectors (see property lattice_vectors). This list indicates if the direction is
            periodic (value 1) or non-periodic (value 0). Note: the elements in this list each
            refer to the direction of the corresponding entry in lattice_vectors and not
            the Cartesian x, y, z directions.
        ''')

    lattice_vectors = Quantity(
        type=np.dtype('f8'), shape=[3, 3], unit=ureg.angstrom,
        links=optimade_links('h.6.2.9'),
        a_optimade=Optimade(query=False, entry=True),
        description='''
            The three lattice vectors in Cartesian coordinates, in ångström (Å).
        ''')

    cartesian_site_positions = Quantity(
        type=np.dtype('f8'), shape=['nsites', 3], unit=ureg.angstrom,
        links=optimade_links('h.6.2.10'),
        a_optimade=Optimade(query=False, entry=True), description='''
            Cartesian positions of each site. A site is an atom, a site potentially occupied by
            an atom, or a placeholder for a virtual mixture of atoms (e.g., in a virtual crystal
            approximation).
        ''')

    nsites = Quantity(
        type=int, default=0,
        links=optimade_links('h.6.2.11'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True), description='''
            An integer specifying the length of the cartesian_site_positions property.
        ''')

    species_at_sites = Quantity(
        type=str, shape=['nsites'],
        links=optimade_links('h.6.2.12'),
        a_optimade=Optimade(query=False, entry=True), description='''
            Name of the species at each site (where values for sites are specified with the same
            order of the cartesian_site_positions property). The properties of the species are
            found in the species property.
        ''')

    # TODO assemblies

    structure_features = Quantity(
        type=MEnum(['disorder', 'unknown_positions', 'assemblies']), shape=['1..*'],
        links=optimade_links('h.6.2.15'),
        a_search=Search(),
        a_optimade=Optimade(query=True, entry=True), description='''
            A list of strings that flag which special features are used by the structure.

            - disorder: This flag MUST be present if any one entry in the species list has a
            chemical_symbols list that is longer than 1 element.
            - unknown_positions: This flag MUST be present if at least one component of the
            cartesian_site_positions list of lists has value null.
            - assemblies: This flag MUST be present if the assemblies list is present.
        ''')

    species = SubSection(sub_section=Species.m_def, repeats=True)



import re
from typing import Union, Dict, Callable, Any, Tuple, List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from optimade.models import EntryResource


class EntryAdapter:
    """Base class for lazy resource entry adapters
    :param entry: JSON OPTIMADE single resource entry.
    """

    ENTRY_RESOURCE: EntryResource = EntryResource
    _type_converters: Dict[str, Callable] = {}

    def __init__(self, entry: dict):
        self._entry = None
        self._converted = {}

        self.entry = entry

        # Note that these return also the default values for otherwise non-provided properties.
        self._common_converters = {
            "json": self.entry.json,  # Return JSON serialized string, see https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeljson
            "dict": self.entry.dict,  # Return Python dict, see https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeldict
        }

    @property
    def entry(self):
        """Get OPTIMADE entry"""
        return self._entry

    @entry.setter
    def entry(self, value: dict):
        """Set OPTIMADE entry
        If already set, print that this can _only_ be set once.
        """
        if self._entry is None:
            self._entry = self.ENTRY_RESOURCE(**value)
        else:
            print("entry can only be set once and is already set.")

    def convert(self, format: str) -> Any:
        """Convert OPTIMADE entry to desired format"""
        if (
            format not in self._type_converters
            and format not in self._common_converters
        ):
            raise AttributeError(
                f"Non-valid entry type to convert to: {format}\n"
                f"Valid entry types: {tuple(self._type_converters.keys()) + tuple(self._common_converters.keys())}"
            )

        if self._converted.get(format, None) is None:
            if format in self._type_converters:
                self._converted[format] = self._type_converters[format](self.entry)
            else:
                self._converted[format] = self._common_converters[format]()

        return self._converted[format]

    @staticmethod
    def _get_model_attributes(
        starting_instances: Union[Tuple[BaseModel], List[BaseModel]], name: str
    ) -> Any:
        """Helper method for retrieving the OPTIMADE model's attribute, supporting "."-nested attributes"""
        for res in starting_instances:
            nested_attributes = name.split(".")
            for nested_attribute in nested_attributes:
                if nested_attribute in getattr(res, "__fields__", {}):
                    res = getattr(res, nested_attribute)
                else:
                    res = None
                    break
            if res is not None:
                return res
        raise AttributeError

    def __getattr__(self, name: str) -> Any:
        """Get converted entry or attribute from OPTIMADE entry
        Support any level of "."-nested OPTIMADE ENTRY_RESOURCE attributes, e.g., `attributes.species` for StuctureResource.
        NOTE: All nested attributes must individually be subclasses of `pydantic.BaseModel`,
        i.e., one can not access nested attributes in lists by passing a "."-nested `name` to this method,
        e.g., `attributes.species.name` or `attributes.species[0].name` will not work for variable `name`.

        Order:
        - Try to return converted entry if using `as_<_type_converters key>`.
        - Try to return OPTIMADE ENTRY_RESOURCE (nested) attribute.
        - Try to return OPTIMADE ENTRY_RESOURCE.attributes (nested) attribute.
        - Raise AttributeError
        """
        # as_<entry_type>
        if name.startswith("as_"):
            entry_type = "_".join(name.split("_")[1:])
            return self.convert(entry_type)

        # Try returning ENTRY_RESOURCE attribute
        try:
            res = self._get_model_attributes((self.entry, self.entry.attributes), name)
        except AttributeError:
            pass
        else:
            return res

        # Non-valid attribute
        entry_resource_name = re.match(
            r"(<class ')([a-zA-Z_]+\.)*([a-zA-Z_]+)('>)", str(self.ENTRY_RESOURCE)
        )
        entry_resource_name = (
            entry_resource_name.group(3)
            if entry_resource_name is not None
            else "UNKNOWN RESOURCE"
        )
        raise AttributeError(
            f"Unknown attribute: {name}\n"
            "If you want to get a converted entry as <entry_type> use `as_<entry_type>`, "
            f"where `<entry_type>` is one of {tuple(self._type_converters.keys()) + tuple(self._common_converters.keys())}\n"
            f"Otherwise, you can try to retrieve an OPTIMADE {entry_resource_name} attribute or property."
        )



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import ase

from nomad.datamodel.metainfo.public import section_k_band, section_band_gap, section_system, section_brillouin_zone
from nomad.normalizing.normalizer import Normalizer
from nomad import config, atomutils
from nomad.constants import pi


class BandStructureNormalizer(Normalizer):
    """Normalizer with the following responsibilities:

      - Calculates band gap(s) if present (section_band_gap, section_band_gap_spin_up, section_band_gap_spin_down).
      - TODO: Creates labels for special points within the band path (band_segm_labels)
      - TODO: Determines if the path is a standard one or not (is_standard)
    """
    def normalize(self, logger=None) -> None:
        # Setup logger
        if logger is not None:
            self.logger = logger.bind(normalizer=self.__class__.__name__)

        # Do nothing if section_run is not present
        if self.section_run is None:
            return

        # Loop through the bands
        for scc in self.section_run.section_single_configuration_calculation:

            # In order to resolve band gaps, we need a reference to the highest
            # occupied energy (semiconductors/insulators) or the Fermi energy
            # (metals)
            e_valence = scc.energy_reference_highest_occupied
            e_fermi = scc.energy_reference_fermi
            energy_reference = e_fermi if e_valence is None else e_valence

            # In order to resolve the special points and the reciprocal cell,
            # we need information about the system.
            system = scc.single_configuration_calculation_to_system_ref

            for band in scc.section_k_band:
                valid_band = self.validate_band(band)
                if valid_band:
                    self.add_reciprocal_cell(band, system)
                    self.add_brillouin_zone(band)
                    self.add_band_gaps(band, energy_reference)
                    self.add_path_labels(band, system)

    def validate_band(self, band: section_k_band) -> bool:
        """Used to check that a band has all required information for normalization.
        """
        if band.band_structure_kind == "vibrational":
            return False
        if len(band.section_k_band_segment) == 0:
            self.logger.info("Could not normalize band structure as band segments are missing.")
            return False
        for segment in band.section_k_band_segment:
            seg_k_points = segment.band_k_points
            seg_energies = segment.band_energies
            if seg_k_points is None or seg_energies is None:
                self.logger.info("Could not normalize band structure as energies or k points are missing.")
                return False
        return True

    def add_reciprocal_cell(self, band: section_k_band, system: section_system):
        """A reciprocal cell for this calculation. If the original unit cell is
        not a primitive one, then we will use the one given by spglib.

        If the used code is calculating it's own primitive cell, then the
        reciprocal cell used in e.g. band structure calculation might differ
        from the one given by spglib. To overcome this we would need to get the
        exact reciprocal cell used by the calculation or then test for
        different primitive cells
        (https://atztogo.github.io/spglib/definition.html#transformation-to-the-primitive-cell)
        whether the k-point path stays inside the first Brillouin zone.
        """
        try:
            orig_atoms = system.m_cache["representative_atoms"]
            symmetry_analyzer = system.section_symmetry[0].m_cache["symmetry_analyzer"]
            prim_atoms = symmetry_analyzer.get_primitive_system()
        except Exception:
            self.logger.info("Could not resolve reciprocal cell.")
            return

        primitive_cell = prim_atoms.get_cell()
        source_cell = orig_atoms.get_cell()

        volume_primitive = primitive_cell.volume
        volume_source = source_cell.volume
        volume_diff = abs(volume_primitive - volume_source)

        if volume_diff > (0.001)**3:
            recip_cell = primitive_cell.reciprocal() * 1e10
        else:
            recip_cell = source_cell.reciprocal() * 1e10

        band.reciprocal_cell = recip_cell

    def add_brillouin_zone(self, band: section_k_band) -> None:
        """Adds the information needed to display the Brillouin zone for this
        material. This functionality could be put into the GUI directly, with
        the Brillouin zone construction performed from the reciprocal cell.

        The Brillouin Zone is a Wigner-Seitz cell, and is thus uniquely
        defined. It's shape does not depend on the used primitive cell.
        """
        recip_cell = band.reciprocal_cell
        if recip_cell is None:
            self.logger.info("Could not resolve Brillouin zone as reciprocal cell is missing.")
            return

        brillouin_zone_data = atomutils.get_brillouin_zone(recip_cell.magnitude)
        section_bz = band.m_create(section_brillouin_zone)
        section_bz.vertices = brillouin_zone_data["vertices"]
        section_bz.faces = brillouin_zone_data["faces"]

    def get_k_space_distance(self, reciprocal_cell: np.array, point1: np.array, point2: np.array) -> float:
        """Used to calculate the Euclidean distance of two points in k-space,
        given relative positions in the reciprocal cell.

        Args:
            reciprocal_cell: Reciprocal cell.
            point1: The first position in k-space.
            point2: The second position in k-space.

        Returns:
            float: Euclidian distance of the two points in k-space in SI units.
        """
        k_point_displacement = np.dot(reciprocal_cell, point1 - point2)
        k_point_distance = np.linalg.norm(k_point_displacement)

        return k_point_distance

    def add_band_gaps(self, band: section_k_band, energy_reference: np.array) -> None:
        """Given the band structure and an energy reference, calculates the band gap
        separately for all spin channels.
        """
        if energy_reference is None:
            self.logger.info("Could not resolve band gaps as the energy reference is missing.")
            return

        reciprocal_cell = band.reciprocal_cell
        if reciprocal_cell is None:
            self.logger.info("Could not resolve band gaps as reciprocal cell is missing.")
            return

        # Gather the energies and k points from each segment into one big
        # array
        reciprocal_cell = reciprocal_cell.magnitude
        path: np.array = []
        energies: np.array = []
        for segment in band.section_k_band_segment:
            seg_k_points = segment.band_k_points
            seg_energies = segment.band_energies
            seg_energies = seg_energies.magnitude
            seg_energies = np.swapaxes(seg_energies, 1, 2)
            path.append(seg_k_points)
            energies.append(seg_energies)

        path = np.concatenate(path, axis=0)
        energies = np.concatenate(energies, axis=2)

        # Handle spin channels separately to find gaps for spin up and down
        n_channels = energies.shape[0]  # pylint: disable=E1136  # pylint/issues/3139
        for channel in range(n_channels):
            eref = energy_reference.magnitude[channel]
            channel_energies = energies[channel, :, :]
            lower_defined = False
            upper_defined = False
            num_bands = channel_energies.shape[0]
            band_indices = np.arange(num_bands)
            band_minima_idx = channel_energies.argmin(axis=1)
            band_maxima_idx = channel_energies.argmax(axis=1)
            band_minima = channel_energies[band_indices, band_minima_idx]
            band_maxima = channel_energies[band_indices, band_maxima_idx]

            # Add a tolerance to minima and maxima
            band_minima_tol = band_minima + config.normalize.band_structure_energy_tolerance
            band_maxima_tol = band_maxima - config.normalize.band_structure_energy_tolerance

            for band_idx in range(num_bands):
                band_min = band_minima[band_idx]
                band_max = band_maxima[band_idx]
                band_min_tol = band_minima_tol[band_idx]
                band_max_tol = band_maxima_tol[band_idx]

                # If any of the bands band crosses the Fermi level, there is no
                # band gap
                if band_min_tol <= eref and band_max_tol >= eref:
                    break
                # Whole band below Fermi level, save the current highest
                # occupied band point
                elif band_min_tol <= eref and band_max_tol <= eref:
                    gap_lower_energy = band_max
                    gap_lower_idx = band_maxima_idx[band_idx]
                    lower_defined = True
                # Whole band above Fermi level, save the current lowest
                # unoccupied band point
                elif band_min_tol >= eref:
                    gap_upper_energy = band_min
                    gap_upper_idx = band_minima_idx[band_idx]
                    upper_defined = True
                    break

            # If a highest point of the valence band and a lowest point of the
            # conduction band are found, and the difference between them is
            # positive, save the information location and value.
            gap = section_band_gap()

            if lower_defined and upper_defined and gap_upper_energy - gap_lower_energy >= 0:

                # See if the gap is direct or indirect by comparing the k-point
                # locations with some tolerance
                k_point_lower = path[gap_lower_idx]
                k_point_upper = path[gap_upper_idx]
                k_point_distance = self.get_k_space_distance(reciprocal_cell, k_point_lower, k_point_upper)
                is_direct_gap = k_point_distance <= config.normalize.k_space_precision

                gap.type = "direct" if is_direct_gap else "indirect"
                gap.value = float(gap_upper_energy - gap_lower_energy)
                gap.conduction_band_min_k_point = k_point_upper
                gap.conduction_band_min_energy = float(gap_upper_energy)
                gap.valence_band_max_k_point = k_point_lower
                gap.valence_band_max_energy = float(gap_lower_energy)
            else:
                gap.value = 0.0

            # Add section to band
            band.m_add_sub_section(section_k_band.section_band_gap, gap)

    def add_path_labels(self, band: section_k_band, system: section_system) -> None:
        """Adds special high symmmetry point labels to the band path. Only k
        points that land on the special points defined by Setyawan/Curtarolo
        are automatically labeled.
        """
        # If labels are already set by the parser dot nothing.
        for segment in band.section_k_band_segment:
            labels = segment.band_segm_labels
            if labels is not None:
                self.logger.info("Existing band segment labels detected, skipping label detection.")
                return

        # Try to get the required data. Fail if not found.
        try:
            cell = system.lattice_vectors.to("angstrom").magnitude
            reciprocal_cell_trans = band.reciprocal_cell.magnitude.T
            bravais_lattice = system.section_symmetry[0].bravais_lattice
        except Exception:
            self.logger.info("Could not resolve path labels as required information is missing.")
            return

        # Find special points for this lattice. If an error occurs, the labels
        # are simply not written.
        try:
            special_points = self.get_special_points(bravais_lattice, cell)
        except Exception as e:
            self.logger.warning("Could not resolve high-symmetry points for the given simulation cell.", exception=e)
            return

        # Form a contiguous array of k points for faster operations
        special_point_labels = list(special_points.keys())
        special_k_points = np.empty((len(special_points), 3))
        for i, kpt in enumerate(special_points.values()):
            special_k_points[i, :] = kpt
        special_k_points_cartesian = np.dot(special_k_points, reciprocal_cell_trans)

        # Match tolerance in 1/m. Taken from the VASP parser.
        eps = config.normalize.k_space_precision

        # Try to find matches for the special points. We only attempt to match
        # points at the start and end of a segment. Any labels set by the
        # parser are overridden, because one cannot ascertain that those labels
        # are consistent across codes.
        for segment in band.section_k_band_segment:
            start_point_cartesian = np.dot(segment.band_k_points[0], reciprocal_cell_trans)
            end_point_cartesian = np.dot(segment.band_k_points[-1], reciprocal_cell_trans)

            # Calculate distance in cartesian space
            start_index = atomutils.find_match(start_point_cartesian, special_k_points_cartesian, eps)
            end_index = atomutils.find_match(end_point_cartesian, special_k_points_cartesian, eps)

            if start_index is None:
                start_label = ""
            else:
                start_label = special_point_labels[start_index]
            if end_index is None:
                end_label = ""
            else:
                end_label = special_point_labels[end_index]
            segment.band_segm_labels = [start_label, end_label]

    def get_special_points(self, bravais_lattice, cell, eps=1e-4):
        """Return dict of special points.

        The definitions are from a paper by Wahyu Setyawana and Stefano
        Curtarolo::

            http://dx.doi.org/10.1016/j.commatsci.2010.05.010

        bravais_lattice: str
            bravais lattice in Pearson notation.
        cell: 3x3 ndarray
            Unit cell.
        eps: float
            Tolerance for cell-check.
        """
        # Special points that do not depend on lattice parameters. TODO: A lot
        # of the bravais lattice are missing from this implementation that is
        # copied from the VASP parser.
        special_points = {
            'cP': {
                'Γ': [0, 0, 0],
                'M': [1 / 2, 1 / 2, 0],
                'R': [1 / 2, 1 / 2, 1 / 2],
                'X': [0, 1 / 2, 0]
            },
            'cF': {
                'Γ': [0, 0, 0],
                'K': [3 / 8, 3 / 8, 3 / 4],
                'L': [1 / 2, 1 / 2, 1 / 2],
                'U': [5 / 8, 1 / 4, 5 / 8],
                'W': [1 / 2, 1 / 4, 3 / 4],
                'X': [1 / 2, 0, 1 / 2]
            },
            'cI': {
                'Γ': [0, 0, 0],
                'H': [1 / 2, -1 / 2, 1 / 2],
                'P': [1 / 4, 1 / 4, 1 / 4],
                'N': [0, 0, 1 / 2]
            },
            'tP': {
                'Γ': [0, 0, 0],
                'A': [1 / 2, 1 / 2, 1 / 2],
                'M': [1 / 2, 1 / 2, 0],
                'R': [0, 1 / 2, 1 / 2],
                'X': [0, 1 / 2, 0],
                'Z': [0, 0, 1 / 2]
            },
            'oP': {
                'Γ': [0, 0, 0],
                'R': [1 / 2, 1 / 2, 1 / 2],
                'S': [1 / 2, 1 / 2, 0],
                'T': [0, 1 / 2, 1 / 2],
                'U': [1 / 2, 0, 1 / 2],
                'X': [1 / 2, 0, 0],
                'Y': [0, 1 / 2, 0],
                'Z': [0, 0, 1 / 2]
            },
            'hP': {
                'Γ': [0, 0, 0],
                'A': [0, 0, 1 / 2],
                'H': [1 / 3, 1 / 3, 1 / 2],
                'K': [1 / 3, 1 / 3, 0],
                'L': [1 / 2, 0, 1 / 2],
                'M': [1 / 2, 0, 0]
            }
        }

        cellpar = ase.geometry.cell_to_cellpar(cell=cell)
        abc = cellpar[:3]
        angles = cellpar[3:] / 180 * pi
        a, b, c = abc
        alpha, _, gamma = angles

        # Check that the unit cells are as in the Setyawana-Curtarolo paper:
        if bravais_lattice == 'cP':
            assert abc.ptp() < eps and abs(angles - pi / 2).max() < eps
        elif bravais_lattice == 'cF':
            assert abc.ptp() < eps and abs(angles - pi / 3).max() < eps
        elif bravais_lattice == 'cI':
            angle = np.arccos(-1 / 3)
            assert abc.ptp() < eps and abs(angles - angle).max() < eps
        elif bravais_lattice == 'tP':
            assert abs(a - b) < eps and abs(angles - pi / 2).max() < eps
        elif bravais_lattice == 'oP':
            assert abs(angles - pi / 2).max() < eps
        elif bravais_lattice == 'hP':
            assert abs(a - b) < eps
            assert abs(gamma - pi / 3 * 2) < eps
            assert abs(angles[:2] - pi / 2).max() < eps
        elif bravais_lattice == 'mP':
            sin_alpha = np.sin(alpha)
            cos_alpha = np.cos(alpha)
            assert c >= a and c >= b
            assert alpha < pi / 2
            assert alpha < pi / 2
            assert (np.abs(angles[1:] - pi / 2) < eps).all()
            eta = (1 - b * cos_alpha / c) / (2 * sin_alpha**2)
            nu = 1 / 2 - eta * c * cos_alpha / b

            return {
                'Γ': [0, 0, 0],
                'A': [1 / 2, 1 / 2, 0],
                'C': [0, 1 / 2, 1 / 2],
                'D': [1 / 2, 0, 1 / 2],
                'D1': [1 / 2, 0, -1 / 2],
                'E': [1 / 2, 1 / 2, 1 / 2],
                'H': [0, eta, 1 - nu],
                'H1': [0, 1 - eta, nu],
                'H2': [0, eta, -nu],
                'M': [1 / 2, eta, 1 - nu],
                'M1': [1 / 2, 1 - eta, nu],
                'M2': [1 / 2, eta, -nu],
                'X': [0, 1 / 2, 0],
                'Y': [0, 0, 1 / 2],
                'Y1': [0, 0, -1 / 2],
                'Z': [1 / 2, 0, 0]
            }
        return special_points[bravais_lattice]



from builtins import next
from builtins import object
import ase.io
import logging
logger = logging.getLogger(__name__)


#===============================================================================
class CoordinateReader(object):
    """Used to parse various different atomic coordinate files.

    See the dictionary 'formats' for all the supported formats and a brief
    explanation. Reading is primarily done by ASE, but in some cases own
    implementation must be used. Returns all coordinates as numpy arrays.
    """
    def __init__(self):
        self.formats = {
            "xyz":    CoordinateFormat(".xyz", "The XYZ file format.", self.ase_iread),
            "cif":    CoordinateFormat(".cif", "Crystallographic Information File", self.ase_iread),
            "pdb":    CoordinateFormat(".pdb", "Protein Data Bank", self.ase_iread),
        }

    def iread(self, file_handle, format):
        """Returns an iterator that goes through the given trajectory file one
        configuration at a time.

        Args:
            file_handle: A file object pointing to the coordinate file.
            format: String containing one of the supported formats.
        """
        if not self.check_format_support(format):
            return

        iread_function = self.formats[format].function
        return iread_function(file_handle, format)

    def n_atoms(self, file_handle, format):
        """Read the first configuration of the coordinate file to extract the
        number of atoms in it.
        """
        iterator = self.iread(file_handle, format)
        pos = next(iterator)
        return pos.shape[0]

    def check_format_support(self, format):
        """Check if the given format is supported.
        """

        if format not in self.formats:
            logger.error("The format '{}' is not supported by CoordinateReader.".format(format))
            return False
        else:
            return True

    def ase_iread(self, file_handle, format):
        """Wrapper for ASE's iread function. Returns numpy arrays instead of
        Atoms objects.
        """
        # The newest ASE version found in Github has an iread function.
        # After reading the ASE source code, it seems that the ASE iread does
        # actually read the entire file into memory and the yields the
        # configurations from it. Should be checked at some point.
        def ase_generator(iterator):
            """Used to wrap an iterator returned by ase.io.iread so that it returns
            the positions instead of the ase.Atoms object.
            """
            for value in iterator:
                yield value.get_positions()

        iterator = ase.io.read(file_handle, index=":", format=format)
        return ase_generator(iterator)

    def custom_iread(self, file_handle, format):
        """
        """
        pass


#===============================================================================
class CoordinateFormat(object):
    """Represents a coordinate format.
    """
    def __init__(self, extension, info, function):
        self.extension = extension
        self.info = info
        self.function = function



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@package    nomadcore.nomad_query
@copyright  Copyright 2018+ Fritz Haber Institute of the Max Planck Society,
            Benjamin Regler - Apache 2.0 License
@license    http://www.apache.org/licenses/LICENSE-2.0
@author     Benjamin Regler
@version    2.0.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import sys
import json
import time
import errno
import random

if sys.version_info.major > 2:
    # For Python 3.0 and later
    from urllib.parse import quote, unquote_plus, urlencode
    from urllib.request import urlopen, Request
else:
    # Fall back to Python 2's urllib2
    from urllib import quote, unquote_plus, urlencode
    from urllib2 import urlopen, Request
    

class NomadQueryResult(object):
    """Nomad Query Result class

    @author Benjamin Regler
    """

    def __init__(self, query, response, version=1.0):
        """Constructor.

        Arguments:
            query {dict}    -- Query information, i.e., query filter, context,
                               group_by, endpoint, and URL
            response {dict} -- Response of the Nomad Query API

        Keyword Arguments:
            version {number} -- Version of the Nomad Query data file
                                (default: {1.0})
        """
        self._uri = []
        self._download_url = ''
        self._query = query or {}
        self._timestamp = int(time.time())

        # Load response information
        self._load(response, version)

    def _load(self, response, version):
        """Load response information

        Arguments:
            response {dict} -- Response of the Nomad Query API
            version {float} -- Version of the Nomad Query data file
        """
        # Set version of the Nomad Query data file
        self._version = version

        # Initialize
        if version == 1.0:
            self._response = response.get('result', {})

            # Construct download path
            path = response.get('path', '')
            self._download_url = self._query.get('endpoint', '') + \
                'download/' + path.split('_')[-1] + '?file=' + \
                quote(path.encode('utf-8')) + '.json'

            # Get Nomad URIs
            response = NomadQuery.request(self._download_url)
            if response['status'] == 'success':
                regex = re.compile(r'(?<=/[a-zA-Z0-9\-_]{3}/)[^.]+')
                paths = response['data'].get('result', [])

                for path in paths:
                    match = regex.search(path)
                    if match:
                        # Substitute prefixes
                        groups = match.group(0).split('/')
                        groups[0] = 'N' + groups[0][1:]         # Normalized

                        if len(groups) == 2:
                            groups[1] = 'C' + groups[1][1:]     # Computed

                        self._uri.append('nmd://' + '/'.join(groups))

        elif version == 2.0:
            self._response = response.get('data', {})

            # Construct and get Nomad URIs
            for entry in self._response:
                if not entry['type'].lower().endswith('calculation'):
                    continue

                # Get archive gid
                context = entry['attributes']['metadata']['archive_context']
                gid = context['archive_gid'][0]

                # Assemble Nomad Uri
                uri = 'nmd://N' + gid[1:] + '/' + entry['id']
                self._uri.append(uri)

    def version(self):
        """Get the version of the Nomad Query data file.

        Returns:
            float -- Version of the Nomad Query data file
        """
        return self._version

    def timestamp(self):
        """Get the timestamp of the query.

        Returns:
            tuple -- The timestamp of the query
        """
        return self._timestamp

    def download_url(self):
        """Get the download URL of the query.

        Returns:
            str -- The download URL of the query

        Deprecated:
            Since version 2.0.0, this method is no longer used by internal code
            and not recommended.
        """
        return self._download_url

    def query(self):
        """Get query information

        Returns:
            dict -- Query information
        """
        return self._query

    def response(self):
        """Get the query response.

        Returns:
            dict -- The query response
        """
        return self._response

    def uri(self):
        """Get Nomad URIs.

        Returns:
            list -- List of Nomad URIs
        """
        return self._uri


class NomadQuery(object):
    """Nomad Query class for accessing stored queries

    @author Benjamin Regler
    """

    # Version of the Nomad Query API
    __version__ = 2.0

    # Private user path
    user_path = '/data/private'

    # Nomad API endpoints
    endpoint = 'https://analytics-toolkit.nomad-coe.eu/api/'
    query_endpoint = 'https://analytics-toolkit.nomad-coe.eu/archive/nql-api/'
    
    def __init__(self, username='', endpoint='', query_endpoint=''):
        """Constructor.

        Keyword Arguments:
            username {str}       -- Current username. Leave empty to auto-detect
                                    username (default: {''})
            endpoint {str}       -- Endpoint of the Nomad API (default:
                                    {'https://analytics-toolkit.nomad-coe.eu/api/'})
            query_endpoint {str} -- Endpoint of the Nomad Query API (default:
                                    {'https://analytics-toolkit.nomad-coe.eu/nql-api/'})
        """
        self._username = ''
        self._base_path = ''

        # Guess username (more like a HACK)
        if not username:
            if os.path.exists(self.user_path):
                paths = os.listdir(self.user_path)
                if len(paths) == 1 and paths[0].lower() != 'nomad':
                    username = paths[0]

        # Set username and overwrite endpoints, if required
        self.username(username)
        if endpoint:
            self.endpoint = str(endpoint)

        if query_endpoint:
            self.query_endpoint = str(query_endpoint)

    def username(self, username=''):
        """Get or set the username.

        Keyword Arguments:
            username {str} -- Current username (default: {''})

        Returns:
            str -- The current username
        """
        if username:
            self._username = str(username)
            self._base_path = os.path.join(self.user_path, self._username,
                                           'nomad-query')
        return self._username

    def resolve(self, nmd, path='', recursive=True, timeout=10):
        """Resolve a Nomad URI.

        A Nomad URI is a base32 encoded string usually 28-characters wide long.
        If separated with a slash, the first URI represents the repository and
        the second URI represents the calculation, e.g.,

            nmd://N2gHWfGLtvdPhqTq8jL6wq1GeLV7_/C093CpZoWf8U32dmRFkp17gy4BCvT .

        Arguments:
            nmd {str} -- Nomad URI with prefix `nmd://`

        Keyword Arguments:
            path {str}       -- Path for resolving Nomad URI, i.e.,
                                /section_run/0c/section_single_configuration_calculation/0c
                                (default: {''})
            recursive {bool} -- Recursively resolve Nomad URI. This may take
                                longer and uses much more memory
                                (default: {True})
            timeout {number} -- Timeout of the request in seconds (default: {10})

        Returns:
            dict -- A python dictionary with data for the given Nomad URI

        Raises:
            ValueError -- Invalid scheme. Nomad URI must start with "nmd://".
            RuntimeError -- Resolving Nomad URI "..." failed.
        """
        if not nmd.startswith('nmd://'):
            raise ValueError('Invalid scheme. Nomad URI must start with "nmd://".')

        # Construct URL
        url = self.endpoint + 'resolve/' + nmd[6:].strip('/')
        if path:
            url += '/' + path.strip('/')
        if recursive:
            url += '?format=recursiveJson'

        response = self.request(url, timeout=timeout)
        if response['status'] != 'success':
            raise RuntimeError('Resolving Nomad URI "%s" failed.' % nmd)

        return response['data']

    def list(self):
        """List all stored queries.

        Returns:
            list -- A sorted list of stored queries. Index 0 always refers to
                    the latest query. Index 1 to the second latest and so on
                    and so forth
        """
        queries = []
        base_path = os.path.join(self._base_path, 'data')
        if not os.path.isdir(base_path):
            return queries

        # Get all stored queries
        for filename in os.listdir(base_path):
            path = os.path.join(base_path, filename)
            if os.path.isfile(path):
                modified = os.path.getmtime(path)
                name = os.path.splitext(filename)[0].split(' ')[-1]

                # Store some useful information about the query
                queries.append({
                    'name': name,
                    'path': path,
                    'filename': filename,
                    'timestamp': int(modified)
                })

        # Sort queries based on modification time
        queries.sort(key=lambda x: -x['timestamp'])
        return queries

    def query(self, query, group_by='', timeout=10, **kwargs):
        """Query the Nomad Database.

        Arguments:
            query {str} -- The query string (see Nomad API reference)

        Keyword Arguments:
            group_by {str}    -- Group-by field. (default: {''})
            num_results {int} -- Number of calculations to return
                                 (default: {10000})
            num_groups {int}  -- Number of distinct calculation groups to return
                                 (default: {10})
            context {str}     -- Deprecated: Query context. Leave empty to use
                                 `single_configuration_calculation` (default: {''})
            compat {bool}     -- Compatibility mode (default: {True})
            timeout {number}  -- Timeout of the request in seconds (default: {10})

        Returns:
            NomadQueryResult -- The Nomad query result

        Raises:
            RuntimeError -- Connection timed out. The Nomad Analytics API
                            Service is currently unavailable.
            RuntimeError -- Unknown error. Please inform the Nomad team to
                            solve this problem.
        """
        # Construct URL
        url = self.query_endpoint + ('search_grouped' if group_by else 'search')
        params = {
            'source_fields': 'archive_gid',
            'sort_field': 'calculation_gid',
            'num_results': max(min(kwargs.get('num_results', 10000), 10000), 1),
            'format': 'nested'
        }

        # Normalize query - compatibility fallback
        if kwargs.get('compat', True):
            query = self._normalize(query)

        # Add query
        params['query'] = query.strip()
        if group_by:
            params['group_by'] = group_by.strip().lower()
            params['num_groups'] = max(kwargs.get('num_groups', 10), 1)

        # Construct URL
        url += '?' + urlencode(params).replace('+', '%20')

        # Read URL
        response = self.request(url, timeout=timeout)
        if response['status'] != 'success':
            raise RuntimeError(response['message'])

        # Check connection timeout
        response = response['data']
        if response['meta'].get('is_timed_out', False) or \
                response['meta'].get('is_terminated_early', False):
            response['message'] = 'Connection timed out.'

        # Construct Nomad Query response
        query = {
            'endpoint': self.query_endpoint,
            'query': params.get('query', ''),
            'group_by': params.get('group_by', ''),
            'url': url
        }

        return NomadQueryResult(query, response, self.__version__)

    def fetch(self, name_or_index='', resolve=False, **params):
        """Fetch stored query.

        Keyword Arguments:
            name_or_index {int|str} -- The name or the index of the query.
                                       Leave empty to use the latest one
                                       (default: {''})
            resolve {bool}          -- Automatically resolve Nomad URIs
                                       (default: {False})
            params {dict}           -- Parameter passed to resolve method. May
                                       include `size` and `seed` for fetching
                                       only a limited number of URIs chosen by
                                       a given seed (default: {{}})

        Returns:
            dict|bool -- A dictionary with query information, Nomad URIs, and
                         optionally resolved Nomad URIs or False if not found

        Raises:
            KeyError -- Query with name "..." does not exists.
        """
        # Get all queries
        queries = self.list()
        if not queries:
            return False

        # Empty name refers to latest query
        index = name_or_index
        if not name_or_index:
            index = 0

        # Extract path
        path = ''
        if isinstance(index, int):
            # Resolve query index
            path = queries[index].get('path')
        else:
            # Load query with specified name
            for i, query in enumerate(queries):
                if query['name'] == index:
                    path = query['path']
                    index = i
                    break

        # Check if path really exists
        if not os.path.exists(path):
            raise KeyError('Query with name "%s" does not exists.' % name_or_index)

        # Load query - JSON
        with open(path, 'r') as file:
            query = json.load(file, 'utf-8')

        # Add useful variables to query
        query.update({
            'data': {},
            'path': path,
            'name': queries[index].get('name')
        })

        # Resolve Nomad URIs?
        if resolve:
            query['data'] = self._resolve(query['uri'], **params)
        return query

    def save(self, name, query, resolve=False, **params):
        """Save query.

        Arguments:
            name {str}     -- Query name
            query {object} -- Instance of `NomadQueryResult`

        Keyword Arguments:
            resolve {bool}  -- Automatically resolve Nomad URIs
                               (default: {False})
            params {dict}   -- Parameter passed to resolve method. May include
                               `size` and `seed` for fetching only a limited
                               number of URIs chosen by a given seed (default: {{}})

        Returns:
            dict -- A dictionary with query information, Nomad URIs, and
                    optionally resolved data

        Raises:
            RuntimeError -- Query name must be an alphanumeric string and may
                            optionally contain underscores and hyphens.
            RuntimeError -- Query parameter must be a `NomadQueryResult` instance.
        """
        # Check query name
        regex = re.compile(r'[^0-9a-z_\-]+', re.I)
        if not name or regex.search(name):
            raise RuntimeError('Query name must be an alphanumeric string and '
                               'may optionally contain underscores and hyphens.')

        # Check query result
        if not isinstance(query, NomadQueryResult):
            raise RuntimeError('Query parameter must be a `NomadQueryResult` '
                               'instance.')

        # Collect information
        timestamp = query.timestamp()
        now = time.localtime(timestamp)
        filename = time.strftime('%Y-%m-%d %H%M%S ', now) + name + '.json'

        data = {
            'filename': filename,
            'version': self.__version__,
            'query': query.query(),
            'timestamp': timestamp,
            'uri': query.uri()
        }

        # Save file
        path = os.path.join(self._base_path, 'data', filename)

        # Write query result
        self._makedir(os.path.dirname(path))
        with open(path, 'w') as file:
            json.dump(data, file, sort_keys=True, indent=4)

        # Add useful variables to query
        data.update({
            'data': {},
            'path': path,
            'name': name
        })

        # Adjust modification time
        os.utime(path, (timestamp, timestamp))

        # Write log file
        path = os.path.join(self._base_path, 'nomad-query.log')
        url = query.query().get('url', '')

        # Create file and check log size
        if not os.path.exists(path):
            self._makedir(os.path.dirname(self.endpoint.rstrip('/')))

            with open(path, 'w') as file:
                file.write('# This file is auto-generated by the Nomad ' +
                           'Query GUI\n# ' + self.endpoint + '\n\n')

        # Append file
        if url:
            with open(path, 'a') as file:
                file.write(time.strftime('[%Y-%m-%d %H:%M:%S] ', now) +
                           unquote_plus(url) + '\n')

        # Resolve Nomad URIs?
        if resolve:
            data['data'] = self._resolve(data['uri'], **params)
        return data

    @staticmethod
    def request(url, timeout=10):
        """Request a URL

        Arguments:
            url {str} -- The URL of a web address

        Keyword Arguments:
            timeout {number} -- Timeout of the request in seconds (default: {10})

        Returns:
            dict -- A dictionary with success status, response data, or 
                    error message
        """
        # Default request response
        result = {
            'url': url,
            'status': 'error',
            'message': 'Unknown error. Please inform the Nomad team to '
                       'solve this problem.'
        }

        try:
            # Get URL
            response = urlopen(Request(url), timeout=timeout)

            # Check response code
            if response.code != 200:
                raise RuntimeError(result['message'])

            # Read response
            data = json.loads(response.read().decode('utf-8'), 'utf-8')

            # Populate result
            result.pop('message')
            result.update({
                'status': 'success',
                'data': data
            })
        except Exception as exc:
            exc = sys.exc_info()[1]
            response = result.copy()

            # Get error message
            message = exc
            if sys.version_info <= (2, 5) and hasattr(exc, 'message'):
                message = exc.message
            elif hasattr(exc, 'reason'):
                message = exc.reason
            response['message'] = str(message)

            # Fix error message
            if response['message'].endswith('timed out'):
                response['message'] = 'Connection timed out. The Nomad ' + \
                    'Analytics API Service is currently unavailable.'

        # Return result
        return result

    def _normalize(self, query):
        """[Protected] Normalize query syntax

        Arguments:
            query {str} -- The query string (see Nomad API reference)

        Returns:
            str -- The normalized query string
        """
        # Convert nomad query syntax v1 to v2
        if re.search(r'(?<!\\):', query):
            values = re.split('\sand\s', query, 0, re.I)

            # Convert query
            regex = re.compile(r'([^:]+):(.+)')
            for i in range(len(values)):
                match = regex.search(values[i])
                if match:
                    # Make sure strings are properly escaped
                    value = map(str.strip, match.group(2).split(','))
                    value = ','.join((v if v.isdigit()
                                      else '"' + v.strip('\'" ') + '"')
                                     for v in value)

                    # Replace colons with equal symbols
                    values[i] = match.group(1) + ' = ' + value

            # Rebuild query
            query = ' AND '.join(values)           

        return query

    def _resolve(self, paths, size=None, seed=None, **params):
        """[Protected] Resolve Nomad URIs.

        Arguments:
            paths {list}  -- List of Nomad URIs

        Keyword Arguments:
            size {number} -- Total number of URIs to resolve (default: {None})
            seed {number} -- Seed to initialize the internal state of the
                             random number generator for selecting n-th Nomad
                             URIs (default: {None})
            params {dict} -- Parameter passed to resolve method (default: {{}})

        Returns:
            dict -- A dictionary with resolved Nomad URIs
        """
        size = params.pop('size', size)
        seed = params.pop('seed', seed)

        # Truncate number of URIs to specified size
        if size and size > 1:
            state = random.getstate()

            # Set seed of random number generator
            if seed is not None:
                random.seed(seed)

            # Get limited set of Nomad URIs
            paths = random.sample(paths, size)
            random.setstate(state)

        return {path: self.resolve(path, **params) for path in paths}

    def _makedir(self, path):
        """[Protected] Recursively create directory.

        This function recursively creates the directory path without raising an
        exception if it does not exists.

        Arguments:
            path {str} -- The path of the directory

        Returns:
            bool -- Returns True on success, False otherwise
        """
        try:
            os.makedirs(path)
        except OSError as exc:  # Python > 2.5
            if exc.errno != errno.EEXIST or not os.path.isdir(path):
                return False
        return True



#!/usr/bin/env python
# Copyright 2016-2018 Rosendo Valero, Fawzi Mohamed, Danio Brambila
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


# xXXXXxx This file copied from GULP  XXXxxXxxXXXX

from __future__ import print_function
from glob import glob
import re
from optparse import OptionParser
from subprocess import Popen, PIPE
import os
from ase.parallel import world
import platform

p = OptionParser()
opts, args = p.parse_args()

#if len(args) == 0:
testfiles = glob('../../test/examples/gamessus/*.out')
testfiles.sort()

    #tests = {}

    #testfiles.sort()
    #print(testfiles)

    #pat = re.compile('test/outputs/example(\d+)[ab]?.got')

    #for fname in testfiles:
    #    print(fname, pat.match(fname))
    #testfiles.sort(key=lambda fname:
    #               int(pat.match(fname).group(1)))
    #for f in testfiles:
    #    print(f)
#else:
#testfiles = argv

for i, testfile in enumerate(testfiles):
    if i % world.size != world.rank:
        continue
    dirname, basename = os.path.split(testfile)
    #print(testfile)
    #print(basename)
    py = 'python'
    if platform.node() == 'labdev-nomad':
        py = '/labEnv3/bin/python'
    args = [py, 'parser_gamess.py', '--annotate']
    args.append(testfile)
    print(' '.join(args))
    proc = Popen(args, stdout=PIPE)
    txt = proc.stdout.read()
    with open('%s.json' % testfile, 'w') as fd:
        fd.write(txt.decode('ascii'))



# pylint: disable=relative-beyond-top-level
import unittest

from optimade.models import LinksResponse

from ..utils import EndpointTestsMixin


class LinksEndpointTests(EndpointTestsMixin, unittest.TestCase):

    request_str = "/links"
    response_cls = LinksResponse


class IndexLinksEndpointTests(EndpointTestsMixin, unittest.TestCase):

    server = "index"
    request_str = "/links"
    response_cls = LinksResponse



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import os
import math
import json
import string
import h5py
import numpy as np
from abc import ABCMeta, abstractmethod
from io import open
import re

import logging


LOGGER = logging.getLogger(__name__)


class ArchiveSection(object):
    __metaclass__ = ABCMeta
    """Defines a storage independent, dictionary like interface to a section
    inside an archive file with the possibility to do recursive searches and
    indexing.

    To make a storage specific implementation just subclass this and
    define the required methods.
    """
    def __init__(self, data, path, archive):
        """
        Args:
            data: The original data that is wrapped by this ArchiveSection.
            path: The full path to this section.
            archive: The Archive object to which this section belongs to.
        """
        self._data = data
        self._path = path
        self._archive = archive

    def is_filtered(self, name):
        """Used to filter out unnecessary information when recursing the
        information. This unnecessary information includes e.g. the gIndex,
        references, etc.
        """
        filtered = set([
            "gIndex",
            "name",
            "references",
            "type",
        ])
        if name not in filtered:
            return name
        return None

    def get_by_path(self, path):
        """Used to query the ArchiveSection recursively with a simple syntax
        that also allows indexing.

        Examples:
            calculation = archive.get_by_path(
                "{repository_hash}/"
                "{calculation_hash}/"
            )
            cell = calculation.get_by_path(
                "section_run:0/"
                "section_system:0/"
                "simulation_cell"
            )
            section_method = calculation.get_by_path(
                "section_run:0/"
                "section_method:0"
            )
            multiple_sections = calculation.get_by_path(
                "section_run:0/"
                "section_method"
            )

        Args:
            path (string): A string indicating the location of the wanted
                information within the metainfo hierarchy. See the examples.

        Returns:
            One of the following: ArchiveSection, a list of ArchiveSections or a concrete value
            corresponding to the given query,
        """
        parts, _, _ = self.get_path_parts(path)
        current_data = self
        current_path = self._path
        n_parts = len(parts)
        for i_part, part in enumerate(parts):

            # Check that the section has not been deleted
            current_path = "{}/{}".format(current_path, part)
            deleted = self.check_deletions(current_path)
            if deleted:
                raise KeyError("Value for '{}' could not be found".format(current_path))

            current_data = current_data.get_child(part)
            if i_part == n_parts - 1:

                filtered = self.is_filtered(part)
                if filtered is None or filtered != part:
                    raise KeyError("Value for '{}' could not be found".format(current_path))
                return current_data

    def get(self, key, default=None):
        """Try to get a value with the given key, if none found, return the
        given default value.

        Args:
            key (string): The key to look for
            default: The value to return if the key is not found.

        Returns:
            The value for the key if found, otherwise returns the given default
            value.
        """
        # Check that the value has not been deleted
        full_path = "{}/{}".format(self._path, key)
        deleted = self.check_deletions(full_path)
        if deleted:
            return default

        # Check cache
        try:
            value = self.get_from_cache(key)
        except KeyError:
            try:
                value = self.get_by_path(key)
            except KeyError:
                value = default
        return value

    def get_scalar(self, key):
        """Get a single value from within this section. If the requested value
        is not scalar, this will raise a KeyError.

        Args:
            key (string): The key to look for

        Returns:
            A scalar value correspondin to the given key.

        Raises:
            KeyError if the found value is not sized like a scalar.
        """
        value = self[key]
        shape = value.shape
        if shape != (1,):
            raise KeyError(
                "The value for '{}/{}' is not shaped like a scalar value. "
                "Instead it has a shape of '{}'."
                .format(self._path, key, shape)
            )
        return value[0]

    def __setitem__(self, path, value):
        """Used to override or set invalid or missing values in the original
        archive.

        This method will store these values into a temporary location
        dictionary that is separate from the original data.
        """
        if self._archive.use_write_cache:

            # If this key had been deleted in the local cache, remove that mark.
            full_path = "{}/{}".format(self._path, path)
            self._archive._deletions.pop(full_path, None)

            # If the value is not already a numpy array, make it so
            if not isinstance(value, np.ndarray):
                if isinstance(value, (list, tuple)):
                    value = np.array(value)
                else:
                    value = np.array((value,))
            self._archive._overrides[full_path] = value
        else:
            raise ValueError(
                "Writing to the source file is currently disabled. If you want "
                "to write to a local cache, set the 'use_write_cache' attribute"
                " of this Archive to True."
            )

    def __delitem__(self, key):
        if self._archive.use_write_cache:
            if key in self:
                full_path = "{}/{}".format(self._path, key)
                self._archive._deletions[full_path] = True
            else:
                raise KeyError(
                    "A value for the given key '{}' has not been set, so it "
                    "could not be deleted."
                    .format(key)
                )
        else:
            raise ValueError(
                "Deleting from the source file is currently disabled. If you "
                "want to enable a local write cache, set the 'use_write_cache' "
                "attribute of this Archive to True."
            )

    def __getitem__(self, key):
        """Used to get a direct child of this section by name.

        Args:
            key (string): The name of the child entry.

        Returns:
            ArchiveSection or a concrete value.
        """
        # Check that the value has not been deleted
        full_path = "{}/{}".format(self._path, key)
        deleted = self.check_deletions(full_path)

        if deleted:
            raise KeyError("Value for '{}' has not been set.".format(full_path))
        try:
            value = self.get_from_cache(key)
        except KeyError:
            value = self.get_by_path(key)
        return value

    @abstractmethod
    def get_child(self, key):
        """This method is used to query the direct children of the section.

        This function will be called internally by the get_by_path(), get() and
        __getitem__() functions.
        """

    @abstractmethod
    def __len__(self):
        """The number of direct children.

        Returns:
            int: The number of direct children.
        """

    @abstractmethod
    def __contains__(self, key):
        """Test for whether this section contains a value for the given key.
        The key can also use the more complex syntax specified in get_by_path.

        Args:
            key (string): The key to check.

        Returns:
            boolean: Whether the given key has a value associated within this
            section.
        """

    @abstractmethod
    def items(self):
        """Used to return the keys and values of the direct children of this
        section.

        Returns:
            An iterable containing the keys and values of the direct values of
            this section.
        """

    @abstractmethod
    def keys(self):
        """Used to return the keys of the direct children of this section.

        Returns:
            An iterable containing the keys of the direct values of this
            section.
        """

    @abstractmethod
    def values(self):
        """Used to return the direct children of this section.

        Returns:
            An iterable containing the direct values of this section.
        """

    def get_from_cache(self, key):
        """Used to get a value from an internal cache that is defined for the
        Arcvive object to which this section belongs to.

        Args:
            key (string): The key to look for in the cache. The given key is
                relative to the path of this section.
        Returns:
            The value found from the cache, is it is set.
        """
        key = "{}/{}".format(self._path, key)
        value = self._archive._overrides[key]
        return value

    def check_deletions(self, full_path):
        """Used to get a value from an internal cache that is defined for the
        Arcvive object to which this section belongs to.

        Args:
            key (string): The key to look for in the cache. The given key is
                relative to the path of this section.
        Returns:
            The value found from the cache, is it is set.
        """
        value = self._archive._deletions.get(full_path)
        if value is True:
            return True
        return False

    def get_path_parts(self, path):
        """Used to separate the given path into sections, and the section to
        name and an index if specified.

        Args:
            path (string):

        Returns:
            tuple: List of section names containing also the index, a list of
                section names without the index, a list of section indices where
                the index is None if no index was specified.

        Raises:
            ValueError if the queried string has an invalid syntax.
        """
        # Remove trailing and leading slash
        if path.endswith("/"):
            path = path[:-1]
        if path.startswith("/"):
            path = path[1:]

        # Go through the hierarchy
        parts = path.split("/")
        indices = []
        names = []
        n_nones = 0
        for part in parts:
            splitted = part.split(":")
            if len(splitted) == 2:
                names.append(splitted[0])
                indices.append(int(splitted[1]))
            elif len(splitted) == 1:
                names.append(splitted[0])
                indices.append(None)
                n_nones += 1
            else:
                raise ValueError(
                    "The given path '{}' is invalid, as there are too many "
                    "colons within a section name."
                    .format(path)
                )
        if n_nones > 3:
            raise ValueError(
                "The given path contains a section without an index in the "
                "middle of the query. You will have to specify the index of "
                "all sections except the last one."
            )
        return parts, names, indices


class Archive(object):
    __metaclass__ = ABCMeta
    """Defines a storage independent interface to an archive file. To make a
    storage specific implementation just subclass this and define the required
    methods.

    Currently only reading the original Archive file is supported. You can also
    set values, but these values will not be written to the the original file
    but will only be valid during the lifetime of this object.

    Attributes:
        filepath (string): Path to the archive file
        repositories (dict): Contains all the repositories within this Archive.
        calculations (dict): Contains all the calculations within this Archive.
        use_write_cache (boolean): Whether to enable writing to a cache that
            will not persist to the original source file.
        _overrides (dictionary): Contains the values that are set during the
            lifetime of this object. These values will not persists on the
            original file.
        _deletions (dictionary): Contains the keys that are deleted during the
            lifetime of this object. These deletions will not be made to the
            original file.
    """
    def __init__(self, filepath, use_write_cache=False):
        """
        Args:
            filepath (string): Filepath to an archive file.
            use_write_cache (boolean): Whether to enable writing to a cache
                that will not persist to the original source file.
        """
        self.filepath = filepath
        self.repositories = {}
        self.calculations = {}
        self.use_write_cache = use_write_cache
        self._overrides = {}
        self._deletions = {}

    @staticmethod
    def factory(archive_path, use_write_cache=False):
        """A factory method for creating Archive objects based on the file
        type.

        Args:
            archive_path (string): The path to the archive file.

        Returns:
            An Archive subclass that is specific to the filetype of the given
            file.
        """
        extension = archive_path.rsplit(".", 1)[-1]
        if extension == "json":
            return ArchiveJSON(archive_path, use_write_cache)
        elif extension == "h5":
            return ArchiveHDF5(archive_path, use_write_cache)
        else:
            raise ValueError(
                "Unknown archive filetype with extension '{}'."
                .format(extension)
            )

    @abstractmethod
    def setup(self, root):
        """Used to setup the dictionaries that contain the repositories and
        calculations.
        """


class ArchiveHDF5(Archive):
    """Provides an access to a HDF5 archive files.

    Also keeps a local cache for the index data. The same index data is often
    queried multiple times, so caching it improves the access speed. The index
    data is also so small that keeping a local cache of it doesn't consume much
    memory.

    Attributes:
        index_cache (dict): A cache containing the index data for groups and
            datasets.
    """
    def __init__(self, filepath, use_write_cache=False):
        super(ArchiveHDF5, self).__init__(filepath, use_write_cache)
        h5_root = h5py.File(filepath, "r")
        self.index_cache = {}
        self.setup(h5_root)

    def setup(self, root):
        """Used to setup the dictionaries that contain the repositories and
        calculations.
        """
        for repo_name, repo in root.items():
            self.repositories[repo_name] = {}
            for calc_name, calc in repo.items():
                calc_section = self.calculations[calc_name] = ArchiveSectionHDF5(
                    calc,
                    "{}/{}".format(repo_name, calc_name),
                    self,
                    [[0]],
                    0
                )
                self.calculations[calc_name] = calc_section
                self.repositories[repo_name][calc_name] = calc_section


class ArchiveSectionHDF5(ArchiveSection):
    """A section from a HDF5 file. Roughly corresponds to the concept of a
    Group in HDF5.
    """
    BASE64DIGITS = string.ascii_uppercase + string.ascii_lowercase + string.digits + "+" + "/"

    def __init__(self, data, path, archive, index_datas, local_index):
        super(ArchiveSectionHDF5, self).__init__(data, path, archive)
        _, names, indices = self.get_path_parts(path)
        # Here we drop out the indices of the repository and calculation
        # section, as they are "None"
        self._indices = indices[2:]
        self._names = names
        self._index_datas = index_datas
        self._local_index = local_index

    def __len__(self):
        return len(self.keys())

    def is_filtered(self, key):
        # return key
        if key.endswith("-index"):
            return None
        key_without_size = key.rsplit(".", 1)[0]
        if key_without_size.endswith("-v"):
            return key_without_size[:-2]
        else:
            return key

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        return True

    def items(self):
        local_index = 0
        for key, value in self._data.items():
            key_filtered = self.is_filtered(key)
            if key_filtered is None:
                continue
            if isinstance(value, h5py.Group):
                index_datas = self._index_datas[:]
                index_data = self._data.get("{}-index".format(key))
                if index_data is None:
                    index_datas = []
                else:
                    index_datas.append(index_data)
                yield (key_filtered, ArchiveSectionHDF5(
                    value,
                    "{}/{}".format(self._path, key),
                    self._archive,
                    index_datas,
                    local_index)
                )
            else:
                yield (key_filtered, value)
            local_index += 1

    def keys(self):
        for key in self._data.keys():
            key_filtered = self.is_filtered(key)
            if key_filtered is not None:
                yield key_filtered

    def values(self):
        local_index = 0
        for key, value in self._data.items():
            key_filtered = self.is_filtered(key)
            if key_filtered is None:
                continue
            if isinstance(value, h5py.Group):
                index_datas = self._index_datas[:]
                index_data = self._data.get("{}-index".format(key))
                if index_data is None:
                    index_datas = []
                else:
                    index_datas.append(index_data)
                yield ArchiveSectionHDF5(
                    value,
                    "{}/{}".format(self._path, key),
                    self._archive,
                    index_datas,
                    local_index
                )
            else:
                yield value
            local_index += 1

    @property
    def mainfile_uri(self):
        """Used to the the mainfile uri from a section corresponding to a
        calculation.
        """
        numpy_uri = self._data.attrs["mainFileUri"]
        str_uri = numpy_uri[0].decode("utf-8")
        return str_uri

    @property
    def parser_info(self):
        """Used to the parser info from a section corresponding to a
        calculation.
        """
        numpy_uri = self._data.attrs["parserInfo"]
        parser_info_str = numpy_uri[0].decode("utf-8")
        parser_info_dict = json.loads(parser_info_str)
        return parser_info_dict

    def check_indices(self):
        """This function is used to check for checking the
        """

    def get_child(self, path):
        """Used to get a direct child of this section within the HDF5 file.
        """
        # Full name for the object that is being searched.
        child_path = "{}/{}".format(self._path, path)

        # Create an index variable
        child_index = None
        if len(self._indices) == 0:
            indices = [0]
        else:
            indices = self._indices[:]
        parent_index = indices[-1]

        # Create a copy of the index data for the child section that is
        # retrieved.
        index_datas = self._index_datas[:]

        # Separate the section number and name
        splitted = path.rsplit(":", 1)
        if len(splitted) == 2:
            name = splitted[0]
            child_index = int(splitted[1])
            indices.append(child_index)
        elif len(splitted) == 1:
            name = path
            child_index = None

        # Check if the path points to a section or a value. Sections will be
        # found simply with the metaname, but the concrete values have a -v and
        # possibly the shape appended to them
        section = True
        try:
            data = self._data[name]
        except KeyError:
            section = False

        # Get the new index data for this child. First try to get the data from
        # a local cache, if not present, then look up from the HDF5 file.
        if section:
            index_path = "{}/{}-index".format(name, name)
        else:
            index_path = "{}-index".format(name)
        global_index_path = "/".join(self._names) + "/" + index_path
        index_data = self._archive.index_cache.get(global_index_path)
        if index_data is None:
            index_data = self._data.get(index_path)
            if index_data is None:
                index_data = np.array([[0]])
            else:
                index_data = index_data.value
            self._archive.index_cache[global_index_path] = index_data
        index_datas.append(index_data)

        # If a section is requested, either provide the section with the
        # provided index, or then a list of section if index was not specified.
        if section:
            # Get the list of available parent indices
            parent_indices = index_data[:, 0]

            # See which section in the index list have the correct parent,
            # and get the correct child from there
            candidates = np.where(parent_indices == parent_index)[0]
            n_candidates = len(candidates)
            if n_candidates == 0:
                raise KeyError(
                    "Could not find value at path '{}'."
                    .format(child_path)
                )
            if child_index is not None:
                try:
                    child_index = candidates[child_index]
                except IndexError:
                    raise KeyError(
                        "Could not find value at path '{}'."
                        .format(child_path)
                    )

                section = self._data[name]
                data = ArchiveSectionHDF5(
                    section,
                    child_path,
                    self._archive,
                    index_datas,
                    child_index
                )
            else:
                group_list = []
                section = self._data[name]
                for candidate in candidates:

                    # Only add the sections that have not been deleted
                    full_path = "{}/{}:{}".format(self._path, name, candidate)
                    deleted = self.check_deletions(full_path)
                    if not deleted:
                        group = ArchiveSectionHDF5(
                            section,
                            full_path,
                            self._archive,
                            index_datas,
                            candidate
                        )
                        group_list.append(group)
                data = group_list

        # If a dataset is requested, check that the dataset exists for this
        # path and then provide it as a numpy array.
        else:
            # Get the local indices of the parent section and search for them
            # in the first column of the index table. There should only be one
            # entry. The index of the data associated with this entry is
            # indicated by the second column. The dataset name can be decided
            # from the data in the remaining columns, which indicate the
            # dimensions of the data.
            test_index = np.where(index_data[:, 0] == self._local_index)[0]
            if test_index.size == 0:
                raise KeyError(
                    "Could not find value at path '{}'."
                    .format(child_path)
                )

            # This error is currently disabled, because it seems that the
            # metainfo system supports repeating scalar values for one section.
            # if test_index.size > 1:
                # raise ValueError(
                    # "The HDF file contains more than one dataset for the "
                    # "path '{}'. "
                    # .format(child_path)
                # )

            # If the value can have multiple shapes, the values are split into
            # different tables. For each table there is a local index in the
            # second column of the index table that we must use.
            data = []
            for row_i in test_index:
                index_row = index_data[row_i]
                if index_row.shape != (1,):
                    data_index = index_row[1]
                else:
                    data_index = row_i

                # The data name may depend on the shape, and if so, the
                # shape is appended to the name as base64 fields
                data_path = name + "-v"
                index_shape = index_data.shape
                if index_shape[1] > 2:
                    data_path = name + "-v"
                    for dim in index_data[data_index][2:]:
                        base64dim = self.base64convert(dim)
                        data_path += ".{}".format(base64dim)

                i_data = self._data[data_path][data_index]

                # Convert bytestrings to regular strings
                if i_data.dtype == np.object:
                    i_data = np.array([x.decode("utf-8") for x in i_data])

                # Gather scalar values to a 1D list
                if i_data.shape == (1,):
                    data.append(i_data[0])
                else:
                    data.append(i_data)

            # If one object returned, remove the outermost list
            if len(test_index) == 1:
                if data[0].shape == ():
                    data = np.array([data[0]])
                else:
                    data = data[0]
            else:
                data = np.array(data)

        return data

    def base64convert(self, x):
        """Used to convert the given base 10 number to base 64.
        """
        base = 64
        if x == 0:
            return ArchiveSectionHDF5.BASE64DIGITS[0]
        digits = []

        while x:
            digits.append(ArchiveSectionHDF5.BASE64DIGITS[x % base])
            x = math.floor(x/base)

        return "".join(digits)


class ArchiveJSON(Archive):
    """Provides a dicionary like access to JSON archive files with the
    possibility to do recursive searches and indexing.

    This implementation will load the entire JSON file into memory, which might
    become a problem with big files and parallel execution on the same machine.
    """
    def __init__(self, filepath, use_write_cache=False):
        super(ArchiveJSON, self).__init__(filepath, use_write_cache)
        with open(filepath, "r") as fin:

            json_root = json.load(fin)

            # Get the calculation name from filename
            calculation_name = os.path.basename(filepath).split(".", 1)[0]

            # Get the repository name from mainFileUri
            mainfile_uri = json_root["mainFileUri"]
            repository_name = mainfile_uri.split("://", 1)[1]
            repository_name = repository_name.split("/", 1)[0]

            root_section = {
                repository_name: {
                    calculation_name: json_root
                }
            }
            self.setup(root_section)

    def setup(self, root):
        for repo_name, repo in root.items():
            self.repositories[repo_name] = {}
            for calc_name, calc in repo.items():
                calc_section = self.calculations[calc_name] = ArchiveSectionJSON(
                    calc,
                    "{}/{}".format(repo_name, calc_name),
                    self
                )
                self.calculations[calc_name] = calc_section
                self.repositories[repo_name][calc_name] = calc_section


class ArchiveSectionJSON(ArchiveSection):
    """Represents a section inside a JSON-file.
    """
    def __len__(self):
        return len(self._data)

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        return True

    def items(self):
        for key, value in self._data.items():
            if self.is_filtered(key) is None:
                continue
            if isinstance(value, dict):
                yield (key, ArchiveSectionJSON(value, "{}/{}".format(self._path, key), self._archive))
            else:
                # Wrap scalar values inside a numpy array to be consistent with the HDF5
                # Archive.
                if not isinstance(value, list):
                    value = np.array([value])
                else:
                    value = np.array(value)
                yield (key, value)

    def keys(self):
        for key in self._data.keys():
            if self.is_filtered(key) is None:
                continue
            yield key

    def values(self):
        for key, value in self._data.items():
            if self.is_filtered(key) is None:
                continue
            if isinstance(value, dict):
                yield ArchiveSectionJSON(value, "{}/{}".format(self._path, key), self._archive)
            else:
                # Wrap scalar values inside a numpy array to be consistent with the HDF5
                # Archive.
                if not isinstance(value, list):
                    value = np.array([value])
                else:
                    value = np.array(value)
                yield value

    @property
    def mainfile_uri(self):
        """Used to the the mainfile uri from a section corresponding to a
        calculation.
        """
        return self._data["mainFileUri"]

    @property
    def parser_info(self):
        """Used to the the parser info from a section corresponding to a
        calculation.
        """
        return self._data["parserInfo"]

    def get_child(self, path):

        def get_root_sections(name):
            section_found = False
            sections = []
            i_section = 0
            while not section_found:
                try:
                    data = self._data["sections"]["{}-{}".format(name, i_section)]
                    sections.append(data)
                except KeyError:
                    section_found = True
                i_section += 1
            if len(sections) == 0:
                raise KeyError("No sections found within the path '{}'".format(path))
            return sections

        # Full name for the object that is being searched.
        child_path = "{}/{}".format(self._path, path)

        # Separate the name and index
        splitted = path.rsplit(":", 1)
        if len(splitted) == 2:
            name = splitted[0]
            index = int(splitted[1])
        else:
            name = splitted[0]
            index = None

        # Determine if section requested
        is_section = False
        if path.startswith("section"):
            is_section = True
        elif re.match(r'^x_\S+_section', path):
            # code-specific section
            is_section = True

        # If no index specified, try to get as concrete value or as a list of
        # sections
        if index is None:
            if not is_section:
                data = self._data[name]
                # Wrap scalar values inside a numpy array to be consistent with the HDF5
                # Archive.
                if not isinstance(data, list):
                    data = np.array([data])
                else:
                    data = np.array(data)
            else:
                try:
                    sections = self._data[name]
                # If this section is section_run, the children will have the
                # index in the section name directly (only specific to JSON)
                except KeyError:
                    sections = get_root_sections(name)

                data = []

                # Only add the sections that have not been deleted
                for i_section, sec in enumerate(sections):
                    full_path = "{}:{}".format(child_path, i_section)
                    deleted = self.check_deletions(full_path)
                    if not deleted:
                        data.append(ArchiveSectionJSON(sec, full_path, self._archive))

        # If index specified, try to get the specific section.
        elif is_section:
            # The section_run and it's child section are within a separate
            # "sections" dictionary, and have indices directly in the name
            try:
                data = self._data["sections"]["{}-{}".format(name, index)]
            except KeyError:
                data = self._data[name]
                try:
                    data = data[index]
                except IndexError:
                    raise KeyError("No sections found within the path '{}'".format(path))
            data = ArchiveSectionJSON(data, child_path, self._archive)
        else:
            raise ValueError(
                "Indexing within a concrete value is not supported. You "
                "will first have to get the whole data and index it then."
            )

        return data
